import json
import logging
from datetime import datetime
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Set,
    TypedDict,
    Union,
    cast,
)
from uuid import uuid4

from httpx import Timeout
from writerai import Writer
from writerai._exceptions import WriterError
from writerai._response import BinaryAPIResponse
from writerai._streaming import Stream
from writerai._types import Body, Headers, NotGiven, Query
from writerai.resources import FilesResource, GraphsResource
from writerai.types import (
    ApplicationListResponse,
    ApplicationRetrieveResponse,
    ChatCompletion,
    Completion,
    CompletionChunk,
    FileDeleteResponse,
    GraphDeleteResponse,
    GraphRemoveFileFromGraphResponse,
    GraphUpdateResponse,
)
from writerai.types import File as SDKFile
from writerai.types import Graph as SDKGraph
from writerai.types.application_generate_content_params import Input
from writerai.types.applications import (
    ApplicationGenerateAsyncResponse,
    JobCreateResponse,
    JobRetryResponse,
)
from writerai.types.applications.application_graphs_response import ApplicationGraphsResponse
from writerai.types.chat_chat_params import GraphData, ToolChoice
from writerai.types.chat_chat_params import Message as WriterAIMessage
from writerai.types.chat_completion_message import ChatCompletionMessage
from writerai.types.question import Question
from writerai.types.question_response_chunk import QuestionResponseChunk
from writerai.types.shared_params.tool_param import FunctionTool as SDKFunctionTool
from writerai.types.shared_params.tool_param import GraphTool as SDKGraphTool
from writerai.types.shared_params.tool_param import LlmTool as SDKLlmTool

from writer.core import get_app_process


class APIOptions(TypedDict, total=False):
    extra_headers: Optional[Headers]
    extra_query: Optional[Query]
    extra_body: Optional[Body]
    timeout: Union[float, Timeout, None, NotGiven]


class ChatOptions(APIOptions, total=False):
    model: str
    tool_choice: ToolChoice
    tools: Union[
            Iterable[
                Union[SDKGraphTool, SDKFunctionTool, SDKLlmTool]
                ],
            NotGiven
        ]
    logprobs: Union[bool, NotGiven]
    max_tokens: Union[int, NotGiven]
    n: Union[int, NotGiven]
    stop: Union[List[str], str, NotGiven]
    temperature: Union[float, NotGiven]
    top_p: Union[float, NotGiven]


class CreateOptions(APIOptions, total=False):
    model: str
    best_of: Union[int, NotGiven]
    max_tokens: Union[int, NotGiven]
    random_seed: Union[int, NotGiven]
    stop: Union[List[str], str, NotGiven]
    temperature: Union[float, NotGiven]
    top_p: Union[float, NotGiven]


class APIListOptions(APIOptions, total=False):
    after: Union[str, NotGiven]
    before: Union[str, NotGiven]
    limit: Union[int, NotGiven]
    order: Union[Literal["asc", "desc"], NotGiven]


class APIRetrieveJobsOptions(APIOptions, total=False):
    limit: Union[int, NotGiven]
    offset: Union[int, NotGiven]
    status: Union[
        Literal["completed", "failed", "in_progress"],
        NotGiven
        ]


class Tool(TypedDict, total=False):
    type: str


class GraphTool(Tool):
    graph_ids: List[str]
    subqueries: bool
    description: Optional[str]


class FunctionToolParameterMeta(TypedDict):
    type: Union[
        Literal["string"],
        Literal["number"],
        Literal["integer"],
        Literal["float"],
        Literal["boolean"],
        Literal["array"],
        Literal["object"],
        Literal["null"]
        ]
    description: Optional[str]
    required: Optional[bool]


class FunctionTool(Tool):
    callable: Callable
    name: str
    description: Optional[str]
    parameters: Dict[str, FunctionToolParameterMeta]


def create_function_tool(
    callable: Callable,
    name: str,
    parameters: Optional[Dict[str, FunctionToolParameterMeta]] = None,
    description: Optional[str] = None
) -> FunctionTool:
    parameters = parameters or {}
    return FunctionTool(
        type="function",
        callable=callable,
        name=name,
        description=description,
        parameters=parameters
    )


class LLMTool(Tool):
    model: str
    description: str


def _process_completion_data_chunk(choice: CompletionChunk) -> str:
    text = choice.value
    if not text:
        return ""
    if isinstance(text, str):
        return text
    raise ValueError("Failed to retrieve text from completion stream")


def _process_chat_data_chunk(chat_data: ChatCompletion) -> tuple[dict, dict]:
    choices = chat_data.choices
    for entry in choices:
        dict_entry = entry.model_dump()
        if dict_entry.get("delta"):
            delta = cast(dict, dict_entry["delta"])

            # Provide content as empty string in case there is no diff
            delta["content"] = delta["content"] or ""
            return dict_entry, delta
        elif dict_entry.get("message"):
            message = cast(dict, dict_entry["message"])

            # Provide content as empty string in case there is no diff
            message["content"] = message["content"] or ""
            return dict_entry, message
    raise ValueError("Failed to retrieve text from chat stream")


class WriterAIManager:
    """
    Manages authentication for Writer AI functionalities.

    :ivar token: Authentication token for the Writer AI API.
    """

    def __init__(self, token: Optional[str] = None):
        """
        Initializes a WriterAIManager instance.


        :param token: Optional; the default token for API authentication used
        if WRITER_API_KEY environment variable is not set up.
        :raises RuntimeError: If an API key was not provided to initialize
        SDK client properly.
        """
        try:
            self.client = Writer(
                # This is the default and can be omitted
                api_key=token,
            )
        except WriterError:
            raise RuntimeError(
                "Failed to acquire Writer API key. " +
                "Provide it by either setting a WRITER_API_KEY" +
                " environment variable, or by initializing the" +
                " AI module explicitly: writer.ai.init(\"my-writer-api-key\")"
                ) from None
        current_process = get_app_process()
        setattr(current_process, 'ai_manager', self)

    @classmethod
    def acquire_instance(cls) -> 'WriterAIManager':
        """
        Retrieve the existing instance of WriterAIManager from
        the current app process.
        If no instance was previously initialized, creates a new one and
        attaches it to the current app process.

        :returns: The current instance of the manager.
        """
        instance: WriterAIManager
        current_process = get_app_process()

        # If instance was not created explicitly, we initialize a new one
        try:
            instance = getattr(current_process, 'ai_manager')
        except AttributeError:
            instance = cls()
        return instance

    @classmethod
    def authorize(cls, token: str):
        """
        Authorize the WriterAIManager with a new token.
        This can be done as an alternative to setting up an environment
        variable, or to override the token that was already provided before.

        :param token: The new token to use for authentication.
        """
        instance = cls.acquire_instance()
        instance.client = Writer(api_key=token)

    @classmethod
    def use_chat_model(cls) -> str:
        """
        Get the configuration for the chat model.

        :returns: Name for the chat model.
        """
        return "palmyra-x-004"

    @classmethod
    def use_completion_model(cls) -> str:
        """
        Get the configuration for the completion model.

        :returns: Name for the completion model.
        """
        return "palmyra-x-004"

    @classmethod
    def acquire_client(cls) -> Writer:
        instance = cls.acquire_instance()
        return instance.client


class SDKWrapper:
    """
    A wrapper class for SDK objects, allowing dynamic access to properties.

    Attributes:
        _wrapped (Union[SDKFile, SDKGraph]): The wrapped SDK object.
    """
    _wrapped: Union[SDKFile, SDKGraph]

    def _get_property(self, property_name):
        """
        Retrieves a property from the wrapped object.

        :param property_name: The name of the property to retrieve.
        :type property_name: str
        :returns: The value of the requested property.
        :raises AttributeError: If the property does not exist.
        """
        try:
            return getattr(self._wrapped, property_name)
        except AttributeError:
            raise AttributeError(
                f"type object '{self.__class__}' " +
                f"has no attribute {property_name}"
                ) from None


class Graph(SDKWrapper):
    """
    A wrapper class for SDKGraph objects, providing additional functionality.

    Attributes:
        _wrapped (writerai.types.Graph): The wrapped SDK Graph object.
        stale_ids (set): A set of stale graph IDs that need updates.
    """
    _wrapped: SDKGraph
    stale_ids: Set[str] = set()

    def __init__(
            self,
            graph_object: SDKGraph
            ):
        """
        Initializes the Graph with the given SDKGraph object.

        :param graph_object: The SDKGraph object to wrap.
        :type graph_object: writerai.types.Graph
        """
        self._wrapped = graph_object

    @staticmethod
    def _retrieve_graphs_accessor() -> GraphsResource:
        """
        Acquires the graphs accessor from the WriterAIManager
        singleton instance.

        :returns: The graphs accessor instance.
        :rtype: GraphsResource
        """
        return WriterAIManager.acquire_client().graphs

    @property
    def is_stale(self):
        return self.id in Graph.stale_ids

    @property
    def id(self) -> str:
        return self._get_property('id')

    @property
    def created_at(self) -> datetime:
        return self._get_property('created_at')

    def _fetch_object_updates(self, force=False):
        """
        Fetches updates for the graph object if it is stale.
        """
        def _get_fresh_object():
            graphs = self._retrieve_graphs_accessor()
            fresh_object = graphs.retrieve(self.id)
            self._wrapped = fresh_object

        if self.is_stale or force is True:
            _get_fresh_object()
        if self.is_stale:
            Graph.stale_ids.remove(self.id)

    @property
    def name(self) -> str:
        self._fetch_object_updates()
        return self._wrapped.name

    @property
    def description(self) -> Optional[str]:
        self._fetch_object_updates()
        return self._wrapped.description

    @property
    def file_status(self):
        self._fetch_object_updates(force=True)
        return self._wrapped.file_status

    def update(
            self,
            name: Optional[str] = None,
            description: Optional[str] = None,
            config: Optional[APIOptions] = None
            ) -> GraphUpdateResponse:
        """
        Updates the graph with the given parameters.

        :param name: The new name for the graph.
        :type name: Optional[str]
        :param description: The new description for the graph.
        :type description: Optional[str]
        :param config: Additional configuration options.
        :type config: Optional[APIOptions]
        :returns: The response from the update operation.
        :rtype: GraphUpdateResponse

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.
        """
        config = config or {}

        # We use the payload dictionary
        # to distinguish between None-values
        # and NotGiven values
        payload = {}
        if name:
            payload["name"] = name
        if description:
            payload["description"] = description
        graphs = self._retrieve_graphs_accessor()
        response = graphs.update(self.id, **payload, **config)
        Graph.stale_ids.add(self.id)
        return response

    def add_file(
            self,
            file_id_or_file: Union['File', str],
            config: Optional[APIOptions] = None
            ) -> 'File':
        """
        Adds a file to the graph.

        :param file_id_or_file: The file object or file ID to add.
        :type file_id_or_file: Union['File', str]
        :param config: Additional configuration options.
        :type config: Optional[APIOptions]
        :returns: The added file object.
        :rtype: File
        :raises ValueError: If the input is neither a File object
        nor a file ID string.

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.
        """
        config = config or {}
        file_id = None
        if isinstance(file_id_or_file, File):
            file_id = file_id_or_file.id
        elif isinstance(file_id_or_file, str):
            file_id = file_id_or_file
        else:
            raise ValueError(
                "'Graph.add_file' method accepts either 'File' object" +
                f" or ID of file as string; got '{type(file_id_or_file)}'"
                )
        graphs = self._retrieve_graphs_accessor()
        response = graphs.add_file_to_graph(
            graph_id=self.id,
            file_id=file_id,
            **config
            )
        Graph.stale_ids.add(self.id)
        return File(response)

    def remove_file(
            self,
            file_id_or_file: Union['File', str],
            config: Optional[APIOptions] = None
            ) -> Optional[GraphRemoveFileFromGraphResponse]:
        """
        Removes a file from the graph.

        :param file_id_or_file: The file object or file ID to remove.
        :type file_id_or_file: Union['File', str]
        :param config: Additional configuration options.
        :type config: Optional[APIOptions]
        :returns: The response from the remove operation.
        :rtype: Optional[GraphRemoveFileFromGraphResponse]
        :raises ValueError: If the input is neither a File object
        nor a file ID string.

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.
        """
        config = config or {}
        file_id = None
        if isinstance(file_id_or_file, File):
            file_id = file_id_or_file.id
        elif isinstance(file_id_or_file, str):
            file_id = file_id_or_file
        else:
            raise ValueError(
                "'Graph.remove_file' method accepts either 'File' object" +
                f" or ID of file as string; got '{type(file_id_or_file)}'"
                )
        graphs = self._retrieve_graphs_accessor()
        response = graphs.remove_file_from_graph(
            graph_id=self.id,
            file_id=file_id,
            **config
            )
        Graph.stale_ids.add(self.id)
        return response

    def _question(
        self,
        question: str,
        stream: bool = True,
        subqueries: bool = False,
        config: Optional[APIOptions] = None
    ):
        if question == "":
            logging.warning(
                'Using empty `question` string ' +
                'against `graphs.question` resource. ' +
                'The model is not likely to produce a meaningful response.'
                )
        config = config or {}
        graphs = self._retrieve_graphs_accessor()
        response = graphs.question(
                graph_ids=[self.id,],
                question=question,
                subqueries=subqueries,
                stream=stream,
                **config
            )
        return response

    def stream_ask(
            self,
            question: str,
            subqueries: bool = False,
            config: Optional[APIOptions] = None
    ) -> Generator[str, None, None]:
        """
        Streams response for a question posed to the graph.

        This method returns incremental chunks of the response, ideal for long 
        responses or when reduced latency is needed.

        :param question: The query or question to be answered by the graph.
        :param subqueries: Enables subquery generation if set to True,
        enhancing the result.
        :param config: Optional dictionary for additional API
        configuration settings.

        :yields: Incremental chunks of the answer to the question.

        :raises ValueError: If an invalid graph or graph ID
        is provided in `graphs_or_graph_ids`.

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.

        **Example Usage**:

        >>> for chunk in graph.stream_ask(
        ...     question="What are the benefits of renewable energy?"
        ... ):
        ...     print(chunk)
        ...
        """

        response = cast(
                Stream[QuestionResponseChunk],
                self._question(
                    question=question,
                    subqueries=subqueries,
                    stream=True,
                    config=config
                )
            )
        for chunk in response._iter_events():
            raw_data = chunk.data
            answer = ""
            try:
                data = json.loads(raw_data)
                answer = data.get("answer", "")
            except json.JSONDecodeError:
                logging.error(
                    "Couldn't parse chunk data during `question` streaming"
                    )
            yield answer

    def ask(
            self,
            question: str,
            subqueries: bool = False,
            config: Optional[APIOptions] = None
    ):
        """
        Sends a question to the graph and retrieves 
        a single response.

        :param question: The query or question to be answered by the graph.
        :param subqueries: Enables subquery generation if set to True,
        enhancing the result.
        :param config: Optional dictionary for additional API
        configuration settings.

        :return: The answer to the question from the graph(s).

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.

        **Example Usage**:

        >>> response = graph.ask(
        ...     question="What is the capital of France?",
        ... )
        """
        response = cast(
            Question,
            self._question(
                question=question,
                subqueries=subqueries,
                stream=False,
                config=config
                )
            )
        return response.answer


def create_graph(
        name: str,
        description: Optional[str] = None,
        config: Optional[APIOptions] = None
        ) -> Graph:
    """
    Creates a new graph with the given parameters.

    :param name: The name of the graph.
    :type name: str
    :param description: The description of the graph.
    :type description: Optional[str]
    :param config: Additional configuration options.
    :type config: Optional[APIOptions]
    :returns: The created graph object.
    :rtype: Graph

    The `config` dictionary can include the following keys:
    - `extra_headers` (Optional[Headers]):
    Additional headers for the request.
    - `extra_query` (Optional[Query]):
    Additional query parameters for the request.
    - `extra_body` (Optional[Body]):
    Additional body parameters for the request.
    - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
    Timeout for the request in seconds.
    """
    config = config or {}
    graphs = Graph._retrieve_graphs_accessor()
    graph_object = graphs.create(
        name=name,
        description=description or NotGiven(),
        **config
        )
    converted_object = cast(SDKGraph, graph_object)
    graph = Graph(converted_object)
    return graph


def retrieve_graph(
        graph_id: str,
        config: Optional[APIOptions] = None
        ) -> Graph:
    """
    Retrieves a graph by its ID.

    :param graph_id: The ID of the graph to retrieve.
    :type graph_id: str
    :param config: Additional configuration options.
    :type config: Optional[APIOptions]
    :returns: The retrieved graph object.
    :rtype: Graph

    The `config` dictionary can include the following keys:
    - `extra_headers` (Optional[Headers]):
    Additional headers for the request.
    - `extra_query` (Optional[Query]):
    Additional query parameters for the request.
    - `extra_body` (Optional[Body]):
    Additional body parameters for the request.
    - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
    Timeout for the request in seconds.
    """
    config = config or {}
    graphs = Graph._retrieve_graphs_accessor()
    graph_object = graphs.retrieve(graph_id, **config)
    graph = Graph(graph_object)
    return graph


def list_graphs(config: Optional[APIListOptions] = None) -> List[Graph]:
    """
    Lists all graphs with the given configuration.

    :param config: Additional configuration options.
    :type config: Optional[APIListOptions]
    :returns: A list of graph objects.
    :rtype: List[Graph]

    The `config` dictionary can include the following keys:
    - `extra_headers` (Optional[Headers]):
    Additional headers for the request.
    - `extra_query` (Optional[Query]):
    Additional query parameters for the request.
    - `extra_body` (Optional[Body]):
    Additional body parameters for the request.
    - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
    Timeout for the request in seconds.
    - `after` (Union[str, NotGiven]):
    Filter to retrieve items created after a specific cursor.
    - `before` (Union[str, NotGiven]):
    Filter to retrieve items created before a specific cursor.
    - `limit` (Union[int, NotGiven]):
    The number of items to retrieve.
    - `order` (Union[Literal["asc", "desc"], NotGiven]):
    The order in which to retrieve items.
    """
    config = config or {}
    graphs = Graph._retrieve_graphs_accessor()
    sdk_graphs = graphs.list(**config)
    return [Graph(sdk_graph) for sdk_graph in sdk_graphs]


def delete_graph(graph_id_or_graph: Union[Graph, str]) -> GraphDeleteResponse:
    """
    Deletes a graph by its ID or object.

    :param graph_id_or_graph: The graph object or graph ID to delete.
    :type graph_id_or_graph: Union[Graph, str]
    :returns: The response from the delete operation.
    :rtype: GraphDeleteResponse
    :raises ValueError: If the input is neither a Graph object
    nor a graph ID string.
    """
    graph_id = None
    if isinstance(graph_id_or_graph, Graph):
        graph_id = graph_id_or_graph.id
    elif isinstance(graph_id_or_graph, str):
        graph_id = graph_id_or_graph
    else:
        raise ValueError(
            "'delete_graph' method accepts either 'Graph' object" +
            f" or ID of graph as string; got '{type(graph_id_or_graph)}'"
            )
    graphs = Graph._retrieve_graphs_accessor()
    return graphs.delete(graph_id)


class File(SDKWrapper):
    """
    A wrapper class for SDK File objects, providing additional functionality.

    Attributes:
        _wrapped (writerai.types.File): The wrapped SDKFile object.
    """
    _wrapped: SDKFile

    def __init__(self, file_object: SDKFile):
        """
        Initializes the File with the given SDKFile object.

        :param file_object: The SDKFile object to wrap.
        :type file_object: writerai.types.File
        """
        self._wrapped = file_object

    @staticmethod
    def _retrieve_files_accessor() -> FilesResource:
        """
        Acquires the files client from the WriterAIManager singleton instance.

        :returns: The files client instance.
        :rtype: FilesResource
        """
        return WriterAIManager.acquire_client().files

    @property
    def id(self) -> str:
        return self._get_property('id')

    @property
    def created_at(self) -> datetime:
        return self._get_property('created_at')

    @property
    def graph_ids(self) -> List[str]:
        return self._get_property('graph_ids')

    @property
    def name(self) -> str:
        return self._get_property('name')

    def download(self) -> BinaryAPIResponse:
        """
        Downloads the file content.

        :returns: The response containing the file content.
        :rtype: BinaryAPIResponse
        """
        files = self._retrieve_files_accessor()
        return files.download(self.id)


def retrieve_file(file_id: str, config: Optional[APIOptions] = None) -> File:
    """
    Retrieves a file by its ID.

    :param file_id: The ID of the file to retrieve.
    :type file_id: str
    :param config: Additional configuration options.
    :type config: Optional[APIOptions]
    :returns: The retrieved file object.
    :rtype: writerai.types.File

    The `config` dictionary can include the following keys:
    - `extra_headers` (Optional[Headers]):
    Additional headers for the request.
    - `extra_query` (Optional[Query]):
    Additional query parameters for the request.
    - `extra_body` (Optional[Body]):
    Additional body parameters for the request.
    - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
    Timeout for the request in seconds.
    """
    config = config or {}
    files = File._retrieve_files_accessor()
    file_object = files.retrieve(file_id, **config)
    file = File(file_object)
    return file


def list_files(config: Optional[APIListOptions] = None) -> List[File]:
    """
    Lists all files with the given configuration.

    :param config: Additional configuration options.
    :type config: Optional[APIListOptions]
    :returns: A list of file objects.
    :rtype: List[File]

    The `config` dictionary can include the following keys:
    - `extra_headers` (Optional[Headers]):
    Additional headers for the request.
    - `extra_query` (Optional[Query]):
    Additional query parameters for the request.
    - `extra_body` (Optional[Body]):
    Additional body parameters for the request.
    - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
    Timeout for the request in seconds.
    - `after` (Union[str, NotGiven]):
    Filter to retrieve items created after a specific cursor.
    - `before` (Union[str, NotGiven]):
    Filter to retrieve items created before a specific cursor.
    - `limit` (Union[int, NotGiven]):
    The number of items to retrieve.
    - `order` (Union[Literal["asc", "desc"], NotGiven]):
    The order in which to retrieve items.
    """
    config = config or {}
    files = File._retrieve_files_accessor()
    sdk_files = files.list(**config)
    return [File(sdk_file) for sdk_file in sdk_files]


def upload_file(
        data: bytes,
        type: str,
        name: Optional[str] = None,
        config: Optional[APIOptions] = None
        ) -> File:
    """
    Uploads a new file with the given parameters.

    :param data: The file content as bytes.
    :type data: bytes
    :param type: The MIME type of the file.
    :type type: str
    :param name: The name of the file.
    :type name: Optional[str]
    :param config: Additional configuration options.
    :type config: Optional[APIOptions]
    :returns: The uploaded file object.
    :rtype: writerai.types.File

    The `config` dictionary can include the following keys:
    - `extra_headers` (Optional[Headers]):
    Additional headers for the request.
    - `extra_query` (Optional[Query]):
    Additional query parameters for the request.
    - `extra_body` (Optional[Body]):
    Additional body parameters for the request.
    - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
    Timeout for the request in seconds.
    """
    config = config or {}
    files = File._retrieve_files_accessor()

    file_name = name or f"WF-{type}-{uuid4()}"
    content_disposition = f'attachment; filename="{file_name}"'

    # Now calling the upload method with correct types.
    sdk_file = files.upload(
        content=data,
        content_type=type,
        content_disposition=content_disposition
        )
    return File(sdk_file)


def delete_file(
        file_id_or_file: Union['File', str],
        config: Optional[APIOptions] = None
        ) -> FileDeleteResponse:
    """
    Deletes a file by its ID or object.

    :param file_id_or_file: The file object or file ID to delete.
    :type file_id_or_file: Union['File', str]
    :param config: Additional configuration options.
    :type config: Optional[APIOptions]
    :returns: The response from the delete operation.
    :rtype: FileDeleteResponse
    :raises ValueError: If the input is neither a File object
    nor a file ID string.

    The `config` dictionary can include the following keys:
    - `extra_headers` (Optional[Headers]):
    Additional headers for the request.
    - `extra_query` (Optional[Query]):
    Additional query parameters for the request.
    - `extra_body` (Optional[Body]):
    Additional body parameters for the request.
    - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
    Timeout for the request in seconds.
    """
    config = config or {}
    file_id = None
    if isinstance(file_id_or_file, File):
        file_id = file_id_or_file.id
    elif isinstance(file_id_or_file, str):
        file_id = file_id_or_file
    else:
        raise ValueError(
            "'delete_file' method accepts either 'File' object" +
            f" or ID of file as string; got '{type(file_id_or_file)}'"
            )

    files = File._retrieve_files_accessor()
    return files.delete(file_id, **config)


class Conversation:
    """
    Manages messages within a conversation flow with an AI system,
    including message validation, history management, and communication
    with an AI model.

    The Conversation class can be initialized in two ways:
    1. By providing an initial system prompt as a string. This starts a new
       conversation, adding a system message with the provided prompt.
       Example:

        >>> conversation = Conversation(
        ...         "You are a social media expert in the financial industry"
        ...         )

    2. By providing a history of messages as a list. This initializes
       the conversation with existing message data.
       Example:

        >>> history = [
        ...         {"role": "user", "content": "Hello"},
        ...         {"role": "assistant", "content": "Hi, how can I help?"}
        ...         ]
        >>> conversation = Conversation(history)

    The class supports both class-wide configuration, which affects
    the entire conversation, and call-specific configuration,
    which can override or complement the class-wide settings
    for specific method calls.

    :param prompt_or_history: Initial system prompt as a string, or
                              history of messages as a list, used to start
                              a new conversation or to load an existing one.
    :param config: Configuration settings for the conversation.

    The `config` dictionary can include the following keys:
    - `extra_headers` (Optional[Headers]):
    Additional headers for the request.
    - `extra_query` (Optional[Query]):
    Additional query parameters for the request.
    - `extra_body` (Optional[Body]):
    Additional body parameters for the request.
    - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
    Timeout for the request in seconds.
    - `model` (str):
    The model to use for completion.
    - `tool_choice` (ToolChoice):
    Configure how the model will call functions: `auto` will allow the model
    to automatically choose the best tool, `none` disables tool calling.
    You can also pass a specific previously defined function.
    - `logprobs` (Union[bool, NotGiven]):
    Specifies whether to return log probabilities of the output tokens.
    - `tools` (Union[Iterable[Union[SDKGraphTool, SDKFunctionTool, SDKLlmTool]], NotGiven]):
    Tools available for the model to use.
    - `max_tokens` (Union[int, NotGiven]):
    Maximum number of tokens to generate.
    - `n` (Union[int, NotGiven]):
    Number of completions to generate.
    - `stop` (Union[List[str], str, NotGiven]):
    Sequences where the API will stop generating tokens.
    - `temperature` (Union[float, NotGiven]):
    Controls the randomness or creativity of the model's responses.
    A higher temperature results in more varied and less predictable text,
    while a lower temperature produces more deterministic
    and conservative outputs.
    - `top_p` (Union[float, NotGiven]):
    Sets the threshold for "nucleus sampling," a technique to focus the model's
    token generation on the most likely subset of tokens. Only tokens with
    cumulative probability above this threshold are considered, controlling the
    trade-off between creativity and coherence.

    **Configuration Example:**

    When initializing, you might provide a general configuration:

    >>> config = {'max_tokens': 100, 'temperature': 0.5}

    >>> conversation = Conversation("Initial prompt", config=config)

    Later, when calling `complete` or `stream_complete`, you can override
    or extend the initial configuration:

    >>> response = conversation.complete(
    ...             config={
    ...                 'max_tokens': 150,
    ...                 'temperature': 0.7
    ...                 }
    ...             )

    This would increase the `max_tokens` limit to 150 and adjust
    the `temperature` to 0.7 for this specific call.

    """
    class Message(TypedDict, total=False):
        """
        Typed dictionary for conversation messages.

        :param role: Specifies the sender role.
        :param content: Text content of the message.
        :param actions: Optional dictionary containing actions
        related to the message.
        """
        role: Literal["system", "assistant", "user", "tool"]
        content: str
        actions: Optional[dict]
        name: Optional[str]
        tool_call_id: Optional[str]
        tool_calls: Optional[List[Dict[str, Union[int, Dict]]]]

    @classmethod
    def validate_message(cls, message):
        """
        Validates if the provided message dictionary matches
        the required structure and values.

        :param message: The message to validate.
        :raises ValueError: If the message structure is incorrect
        or values are inappropriate.
        """
        if not isinstance(message, dict):
            raise ValueError(
                "Attempted to add a non-dict object" +
                f"to the Conversation: {message}"
                )

        if not ("role" in message and "content" in message):
            raise ValueError(
                f"Improper message format to add to Conversation: {message}"
                )

        if not (
            isinstance(message["content"], str)
            or
            message["content"] is None
        ):
            raise ValueError(
                f"Non-string content in message cannot be added: {message}"
                )

        if message["role"] not in ["system", "assistant", "user", "tool"]:
            raise ValueError(f"Unsupported role in message: {message}")

    def __init__(
            self,
            prompt_or_history: Optional[
                Union[str, List['Conversation.Message']]
                ] = None,
            config: Optional[ChatOptions] = None
            ):
        """
        Initializes a new conversation. Two options are possible:

        1. With a system prompt.

        :param system_prompt: The initial message from the system
        to start the conversation.
        :param config: Optional configuration settings for the conversation.

        Example:
        >>> conversation = Conversation(
        ...     "You are a social media expert in the financial industry"
        ...     )

        2. With a history of past messages.

        :param history_import: A list of messages that form the history
        of the conversation.
        :param config: Optional configuration settings for the conversation.

        Example:
        >>> history = [
        ...     {"role": "user", "content": "Hello"},
        ...     {"role": "assistant", "content": "Hi, how can I help?"}
        ...     ]
        >>> conversation = Conversation(history)
        """
        self.messages: List['Conversation.Message'] = []
        if isinstance(prompt_or_history, str):
            # Working with a prompt: adding a system message to history
            prompt = prompt_or_history
            self.add("system", prompt)
        elif isinstance(prompt_or_history, list):
            # Working with a history: trying to add messages
            history_import = prompt_or_history
            for message in history_import:
                self.validate_message(message)
                self += message

        self.config = config or {}
        self._callable_registry: Dict = {}
        self._ongoing_tool_calls: Dict = {}

    def _merge_chunk_to_last_message(self, raw_chunk: dict):
        """
        Merge a chunk of data into the last message in the conversation.

        This method takes a chunk of data and integrates it into the content
        of the last message of the conversation. It appends additional content
        if present, and merges other key-value pairs into the last message's
        dictionary.

        :param raw_chunk: A dictionary containing the chunk of data
        to be merged.
        :raises ValueError: If the Conversation's `messages` list is empty,
        indicating there is no message to merge the chunk with.
        """
        def _clear_chunk_flag(chunk):
            return {
                key: value
                for key, value in chunk.items()
                if key != "chunk"
                }

        if not self.messages:
            raise ValueError("No message to merge chunk with")
        clear_chunk = _clear_chunk_flag(raw_chunk)
        updated_last_message: 'Conversation.Message' = self.messages[-1]
        if "content" in clear_chunk:
            updated_last_message["content"] += clear_chunk.pop("content") or ""

        if "tool_calls" in clear_chunk:
            # Ensure 'tool_calls' exists in updated_last_message as list
            if (
                "tool_calls" not in updated_last_message
                or updated_last_message["tool_calls"] is None
            ):
                updated_last_message["tool_calls"] = []
            # Get the list of new tool calls
            new_tool_calls = clear_chunk.pop("tool_calls", []) or []

            for new_call in new_tool_calls:
                index = new_call.get("index", 0)
                last_message_tool_calls = \
                    cast(list, updated_last_message["tool_calls"])
                if index < len(last_message_tool_calls):
                    # If there's an existing call at this index, update it
                    existing_call = last_message_tool_calls[index]
                    for key, value in new_call.items():
                        if key == "function":
                            if key not in existing_call:
                                existing_call[key] = {}
                            call_function = existing_call[key]
                            # Iterate over function keys and values
                            for fkey, fvalue in value.items():
                                if fkey == "arguments":
                                    # Ensure "arguments" are present
                                    # in existing call as string
                                    if (
                                        fkey not in call_function
                                        or
                                        call_function[fkey] is None
                                    ):
                                        call_function[fkey] = ""

                                    # Append the arguments value
                                    # or an empty string
                                    # to existing arguments
                                    call_function[fkey] += \
                                        fvalue or ""
                                elif fvalue is not None:
                                    call_function[fkey] = fvalue
                        elif value is not None:
                            existing_call[key] = value
                else:
                    # If no existing call, append the new call
                    last_message_tool_calls.append(new_call)

        updated_last_message |= clear_chunk

    @staticmethod
    def _prepare_message(message: 'Conversation.Message') -> WriterAIMessage:
        """
        Converts a message object stored in Conversation to a Writer AI SDK
        `Message` model, suitable for calls to API.

        :param raw_chunk: The data to be merged into the last message.
        :raises ValueError: If there are no messages in the conversation
        to merge with.
        """
        if not ("role" in message and "content" in message):
            raise ValueError("Improper message format")
        sdk_message = WriterAIMessage(
            content=message.get("content", None) or "",
            role=message["role"]
            )
        if msg_name := message.get("name"):
            sdk_message["name"] = cast(str, msg_name)
        if msg_tool_call_id := message.get("tool_call_id"):
            sdk_message["tool_call_id"] = cast(str, msg_tool_call_id)
        if msg_tool_calls := message.get("tool_calls"):
            sdk_message["tool_calls"] = cast(list, msg_tool_calls)
        if msg_graph_data := message.get("graph_data"):
            sdk_message["graph_data"] = cast(
                GraphData,
                msg_graph_data
                )
        if msg_refusal := message.get("refusal"):
            sdk_message["refusal"] = cast(str, msg_refusal)
        return sdk_message

    def _register_callable(
            self,
            callable_to_register: Callable,
            name: str,
            parameters: Dict[str, FunctionToolParameterMeta]
    ):
        """
        Internal helper function to store a provided callable
        for function call, to retrieve it when processing LLM response
        """
        if not callable(callable_to_register):
            raise ValueError(
                f"Expected `{name}` to be callable " +
                f"but got {type(callable_to_register)}"
                )
        self._callable_registry[name] = \
            {
                "callable": callable_to_register,
                "parameters": parameters
            }

    def _clear_callable_registry(self):
        """
        Clear callable registry after LLM response
        """
        self._callable_registry = {}

    def _clear_ongoing_tool_calls(self):
        """
        Clear ongoing tool calls after they've been processed
        """
        self._ongoing_tool_calls = {}

    def _clear_tool_calls_helpers(self):
        self._clear_callable_registry()
        self._clear_ongoing_tool_calls()

    @property
    def _tool_calls_ready(self):
        calls_map = {
            index: "res" in ongoing_tool_call
            for index, ongoing_tool_call
            in self._ongoing_tool_calls.items()
            }
        return all(calls_map.values()) if calls_map else False

    def _gather_tool_calls_messages(self):
        return {
            index: ongoing_tool_call.get(
                "res",
                {
                    "role": "tool",
                    "content": "ERROR: Failed to get function call result – " +
                    "the function was never called. The most likely reason " +
                    "is LLM never issuing a `finish_reason: 'tool_calls'`. " +
                    "Please DO NOT RETRY the function call and inform " +
                    "the user about the error."
                }
            )
            for index, ongoing_tool_call
            in self._ongoing_tool_calls.items()
            }

    def _gather_tool_calls_results(self):
        return list(self._gather_tool_calls_messages().values())

    def _prepare_tool(
            self,
            tool_instance: Union['Graph', GraphTool, FunctionTool, LLMTool]
            ) -> Union[SDKGraphTool, SDKFunctionTool, SDKLlmTool]:
        """
        Internal helper function to process a tool instance
        into the required format.
        """
        def validate_parameters(
                parameters: Dict[str, FunctionToolParameterMeta]
        ) -> bool:
            """
            Validates the `parameters` dictionary to ensure that each key
            is a parameter name, and each value is a dictionary containing
            at least a `type` field, and optionally a `description`.

            :param parameters: The parameters dictionary to validate.
            :return: True if valid, raises ValueError if invalid.
            """
            if not isinstance(parameters, dict):
                raise ValueError("`parameters` must be a dictionary")

            for param_name, param_info in parameters.items():
                if not isinstance(param_info, dict):
                    raise ValueError(
                        f"Parameter '{param_name}' must be a dictionary"
                        )

                if "type" not in param_info:
                    raise ValueError(
                        f"Parameter '{param_name}' must include a 'type' field"
                        )

                if not isinstance(param_info["type"], str):
                    raise ValueError(
                        f"'type' for parameter '{param_name}' must be a string"
                        )
                
                supported_types = {
                    "string", "number", "integer", "float", 
                    "boolean", "array", "object", "null"
                }
                if param_info["type"] not in supported_types:
                    raise ValueError(
                        f"Unsupported type '{param_info['type']}' " +
                        f"for parameter '{param_name}'"
                    )

                # Optional 'description' validation (if provided)
                if (
                    "description" in param_info
                    and
                    not isinstance(param_info["description"], str)
                ):
                    raise ValueError(
                        f"'description' for parameter '{param_name}' " +
                        "must be a string if provided"
                        )

            return True

        def prepare_parameters(parameters: Dict[str, FunctionToolParameterMeta]) -> Dict:
            """
            Prepares the parameters dictionary for a function tool.

            :param parameters: The parameters dictionary to prepare.
            :return: The processed parameters dictionary.
            """
            processed_params: Dict[str, FunctionToolParameterMeta] = {}
            if not parameters:
                return processed_params
            else:
                required_list = []
                for param_name, param_info in parameters.items():
                    processed_param = param_info.copy()
                    # Convert Python numeric types to JSON schema "number" type
                    if processed_param["type"] in ["float", "integer"]:
                        processed_param["type"] = "number"
                    # Check the "required" flag on parameter
                    if processed_param.get("required", False) is True:
                        required_list.append(param_name)
                    processed_params[param_name] = processed_param

            res = {
                "type": "object",
                "properties": processed_params
            }
            if required_list:
                res["required"] = required_list
            return res

        def validate_graph_ids(graph_ids: List[str]) -> bool:
            """
            Validates that `graph_ids` is a list of strings.

            :param graph_ids: The list of graph IDs to validate.
            :return: True if valid, raises ValueError if invalid.
            """
            if not isinstance(graph_ids, list):
                raise ValueError("`graph_ids` must be a list")

            for graph_id in graph_ids:
                if not isinstance(graph_id, str):
                    raise ValueError(f"Graph ID '{graph_id}' must be a string")

            return True

        if isinstance(tool_instance, Graph):
            # Prepare a single graph tool
            return cast(
                SDKGraphTool,
                {
                    "type": "graph",
                    "function": {
                        "graph_ids": [tool_instance.id],
                        "subqueries": True,
                        "description": tool_instance.description
                    }
                }
            )

        elif isinstance(tool_instance, dict):
            # Handle a dictionary (either a graph, a function or a LLM tool)
            if "type" not in tool_instance:
                raise ValueError(
                    "Invalid tool definition: 'type' field is missing"
                    )

            tool_type = tool_instance["type"]

            if tool_type == "graph":
                tool_instance = cast(GraphTool, tool_instance)
                if "graph_ids" not in tool_instance:
                    raise ValueError("Graph tool must include 'graph_ids'")
                if "description" not in tool_instance:
                    logging.warning(
                        "No description provided for `graph` tool. " +
                        "This may produce suboptimal results. " +
                        "To increase output quality, provide a description " +
                        "for the tool."
                        )
                # Return graph tool JSON

                graph_ids_valid = \
                    validate_graph_ids(tool_instance["graph_ids"])
                if graph_ids_valid:
                    return cast(
                        SDKGraphTool,
                        {
                            "type": "graph",
                            "function": {
                                "graph_ids": tool_instance["graph_ids"],
                                "subqueries": tool_instance.get(
                                    "subqueries", False
                                    ),
                                "description": tool_instance.get("description", None)
                            }
                        }
                    )
                else:
                    raise ValueError(
                        "Invalid Graph IDs provided: " +
                        f"{tool_instance['graph_ids']}"
                        )

            elif tool_type == "function":
                tool_instance = cast(FunctionTool, tool_instance)
                if "callable" not in tool_instance:
                    raise ValueError("Function tool missing `callable`")
                if (
                    "name" not in tool_instance
                    or
                    "parameters" not in tool_instance
                ):
                    raise ValueError(
                        "Function tool must include 'name' and 'parameters'"
                        )
                if "description" not in tool_instance:
                    logging.warning(
                        "No description provided for `function` tool. " +
                        "This may produce suboptimal results. " +
                        "To increase output quality, provide a description " +
                        "for the tool."
                        )

                parameters_valid = \
                    validate_parameters(tool_instance["parameters"])
                # Return function tool JSON
                if parameters_valid:
                    self._register_callable(
                        tool_instance["callable"],
                        tool_instance["name"],
                        tool_instance["parameters"]
                        )
                    return cast(
                        SDKFunctionTool,
                        {
                            "type": "function",
                            "function": {
                                "name": tool_instance["name"],
                                "parameters":
                                    prepare_parameters(
                                        tool_instance["parameters"]
                                        ),
                                "description": tool_instance.get("description")
                                }
                        }
                    )
                else:
                    raise ValueError(
                        "Invalid parameters for function " +
                        f"`{tool_instance['name']}`"
                        )

            elif tool_type == "llm":
                tool_instance = cast(LLMTool, tool_instance)
                if "model" not in tool_instance:
                    raise ValueError("LLM tool must include 'model'")
                if "description" not in tool_instance:
                    logging.warning(
                        "No description provided for `llm` tool. " +
                        "This may produce suboptimal results. " +
                        "To increase output quality, provide a description " +
                        "for the tool."
                        )
                # Return LLM tool JSON
                return cast(
                    SDKLlmTool,
                    {
                        "type": "llm",
                        "function": {
                            "model": tool_instance["model"],
                            "description": tool_instance.get("description")
                        }
                    }
                )
            else:
                raise ValueError(f"Unsupported tool type: {tool_type}")

        else:
            raise ValueError(f"Invalid tool input: {tool_instance}")

    def __add__(self, chunk_or_message: Union['Conversation.Message', dict]):
        """
        Adds a message or appends a chunk to the last message
        in the conversation.

        :param chunk_or_message: Dictionary representation of a message
        or chunk to add.
        :raises TypeError: If passed chunk_or_message is not a dictionary.
        :raises ValueError: If chunk_or_message is not a proper message
        with "role" and "content".
        """
        if not isinstance(chunk_or_message, dict):
            raise TypeError(
                "Conversation only supports dict operands for addition"
                )
        if chunk_or_message.get("chunk") is True:
            chunk = chunk_or_message
            self._merge_chunk_to_last_message(cast(dict, chunk))
        else:
            message = chunk_or_message
            self.validate_message(message)
            message_to_append = {
                "role": message["role"],
                "content": message["content"],
                "actions": message.get("actions")
                }
            if "tool_calls" in message:
                message_to_append["tool_calls"] = message["tool_calls"]
            if "tool_call_id" in message and message["role"] == "tool":
                message_to_append["tool_call_id"] = message["tool_call_id"]
            self.messages.append(cast(Conversation.Message, message_to_append))
        return self

    def add(
            self,
            role: Literal["system", "assistant", "user", "tool"],
            message: str
            ):
        """
        Adds a new message to the conversation.

        :param role: The role of the message sender.
        :param message: The content of the message.
        """
        self.__add__({"role": role, "content": message})

    def _send_chat_request(
            self,
            request_model: str,
            request_data: ChatOptions,
            stream: bool = False
    ) -> Union[Stream, ChatCompletion]:
        """
        Helper function to send a chat request to the LLM.

        :param request_model: Model to use for the chat.
        :param request_data: Configuration settings for the chat request.
        :param stream: Whether to use streaming mode.
        :return: The response from the LLM, either as
        a Stream or a ChatCompletion object.
        """
        client = WriterAIManager.acquire_client()
        prepared_messages = [
                self._prepare_message(message)
                for message in self.messages
            ]
        logging.debug(
            "Attempting to request a message from LLM: " +
            f"prepared messages – {prepared_messages}, " +
            f"request_data – {request_data}"
            )
        return client.chat.chat(
            messages=prepared_messages,
            model=request_model,
            stream=stream,
            logprobs=request_data.get('logprobs', NotGiven()),
            tools=request_data.get('tools', NotGiven()),
            tool_choice=request_data.get('tool_choice', NotGiven()),
            max_tokens=request_data.get('max_tokens', NotGiven()),
            n=request_data.get('n', NotGiven()),
            stop=request_data.get('stop', NotGiven()),
            temperature=request_data.get('temperature', NotGiven()),
            top_p=request_data.get('top_p', NotGiven()),
            extra_headers=request_data.get('extra_headers'),
            extra_query=request_data.get('extra_query'),
            extra_body=request_data.get('extra_body'),
            timeout=request_data.get('timeout', NotGiven()),
        )

    def _convert_argument_to_type(self, value: Any, target_type: str) -> Any:
        """
        Converts the argument to the specified target type.

        :param value: The value to convert.
        :param target_type: The target type as a string.
        :return: The converted value.
        :raises ValueError: If the value cannot be converted
        to the target type.
        """
        if target_type == "string":
            return str(value)
        elif target_type == "number":
            return float(value)  # as float is more specific
        elif target_type == "integer":
            return int(value)
        elif target_type == "float":
            return float(value)
        elif target_type == "boolean":
            if isinstance(value, str):
                return value.lower() in ["true", "1"]
            return bool(value)
        elif target_type == "array":
            if isinstance(value, str):
                return json.loads(value)  # Assuming JSON string for arrays
            elif isinstance(value, list):
                return value
            else:
                raise ValueError(f"Cannot convert {value} to list.")
        elif target_type == "object":
            if isinstance(value, str):
                return json.loads(value)  # Assuming JSON string for objects
            elif isinstance(value, dict):
                return value
            else:
                raise ValueError(f"Cannot convert {value} to dict.")
        elif target_type == "null":
            return None
        else:
            raise ValueError(f"Unsupported target type: {target_type}")

    def _check_if_arguments_are_required(self, function_name: str) -> bool:
        callable_entry = self._callable_registry.get(function_name)
        if not callable_entry:
            raise RuntimeError(
                f"Tried to check arguments of function {function_name} " +
                "which is not present in the conversation's callable registry."
                )
        callable_parameters = callable_entry.get("parameters")
        return \
            callable_parameters is not None \
            and \
            callable_parameters != {}

    def _execute_function_tool_call(self, index: int) -> dict:
        """
        Executes the function call for the specified tool call index.

        :param index: The index of the tool call in _ongoing_tool_calls.
        :return: The follow-up message to be sent to the LLM.
        """
        function_name = self._ongoing_tool_calls[index]["name"]
        arguments = self._ongoing_tool_calls[index]["arguments"]
        tool_call_id = self._ongoing_tool_calls[index]["tool_call_id"]

        # Parse arguments and execute callable
        try:
            if (
                not arguments
                and
                not self._check_if_arguments_are_required(function_name)
            ):
                parsed_arguments = {}
            else:
                parsed_arguments = json.loads(arguments)
            callable_entry = self._callable_registry.get(function_name)

            if callable_entry:
                param_specs = callable_entry["parameters"]
                # Convert arguments based on registered parameter types
                converted_arguments = {}
                try:
                    for param_name, param_info in param_specs.items():
                        if param_name in parsed_arguments:
                            target_type = param_info["type"]
                            value = parsed_arguments[param_name]
                            converted_arguments[param_name] = \
                                self._convert_argument_to_type(
                                    value,
                                    target_type
                                    )
                        else:
                            raise ValueError(
                                f"Missing required parameter: {param_name}"
                                )

                    func = callable_entry.get("callable")
                    if not func:
                        raise ValueError(
                            f"Misconfigured function {function_name}: " +
                            "no callable provided"
                            )

                    # Call the function with converted arguments
                    func_result = func(**converted_arguments)
                except Exception as e:
                    logging.error(
                        "An error occured during the execution of function " +
                        f"`{function_name}`: {e}"
                    )
                    func_result = \
                        "Function call failed due to an exception – " + \
                        "please DO NOT RETRY the call and inform the user " + \
                        "about the error"

                # Prepare follow-up message with the function call result
                follow_up_message = {
                    "role": "tool",
                    "name": function_name,
                    "tool_call_id": tool_call_id,
                    "content": f"{func_result}"
                }

                return follow_up_message
            else:
                raise ValueError(
                    f"`{function_name}` is not present in callable registry"
                    )

        except json.JSONDecodeError:
            logging.error(
                f"Failed to parse arguments for tool call: {arguments}"
                )

        return {
                    "role": "tool",
                    "name": function_name,
                    "tool_call_id": tool_call_id,
                    "content":
                            "Failed to parse provided arguments " +
                            "for tool call – please DO NOT RETRY " +
                            "the function call and inform the user " +
                            "about the error."
                }

    def _process_tool_call(
            self,
            index,
            tool_call_id,
            tool_call_name,
            tool_call_arguments
            ):
        if index not in self._ongoing_tool_calls:
            self._ongoing_tool_calls[index] = {
                "name": None,
                "arguments": "",
                "tool_call_id": None
            }

        # Capture `tool_call_id` from the message
        if tool_call_id is not None:
            self._ongoing_tool_calls[index]["tool_call_id"] = tool_call_id

        # Capture `name` for function call
        if tool_call_name is not None:
            self._ongoing_tool_calls[index]["name"] = tool_call_name

        # Accumulate arguments across chunks
        if tool_call_arguments is not None:
            if (
                tool_call_arguments.startswith("{")
                and tool_call_arguments.endswith("}")
            ):
                # For cases when LLM "bugs" and returns
                # the whole arguments string as a last chunk
                fixed_chunk = tool_call_arguments.rsplit("{")[-1]
                self._ongoing_tool_calls[index]["arguments"] = \
                    "{" + fixed_chunk
            else:
                # Process normally
                self._ongoing_tool_calls[index]["arguments"] += \
                    tool_call_arguments

        # Check if we have all necessary data to execute the function
        tool_call_id, tool_call_name, tool_call_arguments = \
            self._ongoing_tool_calls[index]["tool_call_id"], \
            self._ongoing_tool_calls[index]["name"], \
            self._ongoing_tool_calls[index]["arguments"]

        tool_call_id_ready = tool_call_id is not None
        tool_call_name_ready = tool_call_name is not None

        # Check whether the arguments are prepared properly -
        # either present in correct format
        # or should not be used due to not being required for the function
        if tool_call_name_ready:
            # Function name is needed to check the function for params
            tool_call_arguments_not_required = \
                (
                    not tool_call_arguments
                    and
                    not self._check_if_arguments_are_required(
                        tool_call_name
                    )
                )
            tool_call_arguments_formatted_properly = \
                (
                    isinstance(
                        tool_call_arguments, str
                    )
                    and
                    tool_call_arguments.endswith("}")
                )
            tool_call_arguments_ready = \
                tool_call_arguments_not_required \
                or \
                tool_call_arguments_formatted_properly
        else:
            tool_call_arguments_ready = False

        if (
            tool_call_id_ready
            and
            tool_call_name_ready
            and
            tool_call_arguments_ready
        ):
            follow_up_message = self._execute_function_tool_call(index)
            if follow_up_message:
                self._ongoing_tool_calls[index]["res"] = follow_up_message

    def _process_tool_calls(self, message: ChatCompletionMessage):
        if message.tool_calls:
            for helper_index, tool_call in enumerate(message.tool_calls):
                index = tool_call.index or helper_index
                tool_call_id = tool_call.id
                tool_call_name = tool_call.function.name
                tool_call_arguments = tool_call.function.arguments

                self._process_tool_call(
                    index,
                    tool_call_id,
                    tool_call_name,
                    tool_call_arguments
                )

    def _process_streaming_tool_calls(self, chunk: Dict):
        tool_calls = chunk["tool_calls"]
        if isinstance(tool_calls, list):
            for tool_call in tool_calls:
                index = tool_call["index"]
                tool_call_id = tool_call["id"]
                tool_call_name = tool_call["function"]["name"]
                tool_call_arguments = tool_call["function"]["arguments"]

                self._process_tool_call(
                    index,
                    tool_call_id,
                    tool_call_name,
                    tool_call_arguments
                )

    def _prepare_received_message_for_history(
            self,
            message: ChatCompletionMessage
            ):
        """
        Prepares a received message for adding to the conversation history.

        :param message: The message to prepare.
        :return: The prepared message.
        """
        raw_message = message.model_dump()
        if not raw_message.get("content"):
            raw_message["content"] = ""
        return raw_message

    def _process_response_data(
            self,
            response_data: ChatCompletion,
            request_model: str,
            request_data: ChatOptions,
            depth=1,
            max_depth=5
            ) -> 'Conversation.Message':
        if depth > max_depth:
            raise RuntimeError(
                "Reached maximum depth " +
                "when processing response data tool calls."
                )
        for entry in response_data.choices:
            message = entry.message
            if message:
                # Handling tool call fragments
                logging.debug(f"Received message – {message}")
                if message.tool_calls is not None:
                    logging.debug(
                        f"Message has tool calls - {message.tool_calls}"
                        )
                    self += self._prepare_received_message_for_history(message)
                    self._process_tool_calls(message)
                    self.messages += self._gather_tool_calls_results()
                    # Send follow-up call to LLM
                    logging.debug("Sending a request to LLM")
                    follow_up_response = cast(
                        ChatCompletion,
                        self._send_chat_request(
                            request_model=request_model,
                            request_data=request_data
                        )
                    )
                    logging.debug(f"Received response – {follow_up_response}")

                    # Call the function recursively
                    # to either process a new tool call
                    # or return the message if no tool calls are requested

                    self._clear_ongoing_tool_calls()
                    return self._process_response_data(
                        follow_up_response,
                        request_model=request_model,
                        request_data=request_data,
                        depth=depth+1
                        )
                else:
                    return cast(Conversation.Message, message.model_dump())
        raise RuntimeError(
            "Failed to acquire proper response " +
            f"for completion from data: {response_data}"
            )

    def _process_stream_response(
            self,
            response: Stream,
            request_model: str,
            request_data: ChatOptions,
            depth=1,
            max_depth=5,
            flag_chunks=False
    ) -> Generator[dict, None, None]:
        if depth > max_depth:
            raise RuntimeError(
                "Reached maximum depth " +
                "when processing response data tool calls."
                )

        for line in response:
            chunk_data, chunk = _process_chat_data_chunk(line)

            if flag_chunks is True:
                # We avoid flagging first chunk
                # to trigger creating a message
                # to append chunks to
                chunk |= {"chunk": True}

            # Handling tool call fragments
            tool_calls_present = chunk.get("tool_calls") is not None
            tool_calls_need_processing = \
                chunk_data.get("finish_reason") == "tool_calls"
            if tool_calls_present or tool_calls_need_processing:
                # Handle tool calls chunks
                if tool_calls_present:
                    self += chunk
                    self._process_streaming_tool_calls(chunk)
                if tool_calls_need_processing:
                    # Send follow-up call to LLM
                    self.messages += self._gather_tool_calls_results()
                    follow_up_response = cast(
                        Stream,
                        self._send_chat_request(
                            request_model=request_model,
                            request_data=request_data,
                            stream=True
                        )
                    )

                    try:
                        self._clear_ongoing_tool_calls()
                        yield from self._process_stream_response(
                            response=follow_up_response,
                            request_model=request_model,
                            request_data=request_data,
                            depth=depth+1,
                            max_depth=max_depth,
                            flag_chunks=False
                        )
                    finally:
                        follow_up_response.close()
            else:
                # Handle regular message chunks
                if chunk.get("content") is not None:
                    if flag_chunks is False:
                        flag_chunks = True
                    yield chunk

    def complete(
            self,
            config: Optional['ChatOptions'] = None,
            tools: Optional[
                Union[
                    Graph,
                    GraphTool,
                    FunctionTool,
                    LLMTool,
                    List[Union[Graph, GraphTool, FunctionTool, LLMTool]]
                    ]  # can be an instance of tool or a list of instances
                ] = None,
            max_tool_depth: int = 5,
            ) -> 'Conversation.Message':
        """
        Processes the conversation with the current messages and additional
        data to generate a response.
        Note: this method only produces AI model output and does not attach the
        result to the existing conversation history.

        :param tools: Optional tools to use for processing.
        :param config: Optional parameters to pass for processing.
        :param max_tool_depth: Maximum depth for tool calls processing.
        :return: Generated message.
        :raises RuntimeError: If response data was not properly formatted
        to retrieve model text.
        """
        config = config or {'max_tokens': 1024}
        if tools is not None and not isinstance(tools, list):
            tools = [tools]

        prepared_tools = [
            self._prepare_tool(tool_instance)
            for tool_instance in (tools or [])
            ]

        request_data: ChatOptions = {**config, **self.config}
        if prepared_tools:
            request_data |= {"tools": prepared_tools}
        request_model = \
            request_data.get("model") or WriterAIManager.use_chat_model()

        response_data: ChatCompletion = cast(
                ChatCompletion,
                self._send_chat_request(
                    request_model=request_model,
                    request_data=request_data
                )
            )

        response = self._process_response_data(
            response_data,
            request_model=request_model,
            request_data=request_data,
            max_depth=max_tool_depth
            )

        # Clear buffer and callable registry for the completed tool call
        self._clear_tool_calls_helpers()

        return response

    def stream_complete(
            self,
            config: Optional['ChatOptions'] = None,
            tools: Optional[
                Union[
                    Graph,
                    GraphTool,
                    FunctionTool,
                    List[Union[Graph, GraphTool, FunctionTool]]
                    ]  # can be an instance of tool or a list of instances
                ] = None,
            max_tool_depth: int = 5
            ) -> Generator[dict, None, None]:
        """
        Initiates a stream to receive chunks of the model's reply.
        Note: this method only produces AI model output and does not attach
        the result to the existing conversation history.

        :param tools: Optional tools to use for processing.
        :param config: Optional parameters to pass for processing.
        :param max_tool_depth: Maximum depth for tool calls processing.
        :yields: Model response chunks as they arrive from the stream.
        """
        config = config or {}
        if tools is not None and not isinstance(tools, list):
            tools = [tools]

        prepared_tools = [
            self._prepare_tool(tool_instance)
            for tool_instance in (tools or [])
            ]

        request_data: ChatOptions = {**config, **self.config}
        if prepared_tools:
            request_data |= {"tools": prepared_tools}
        request_model = \
            request_data.get("model") or WriterAIManager.use_chat_model()

        response: Stream = cast(
            Stream,
            self._send_chat_request(
                request_model=request_model,
                request_data=request_data,
                stream=True
            )
        )

        yield from self._process_stream_response(
            response=response,
            request_model=request_model,
            request_data=request_data,
            max_depth=max_tool_depth
        )

        # Clear buffer and callable registry for the completed tool call
        self._clear_tool_calls_helpers()
        response.close()

    def _is_serialized(self, message: 'Conversation.Message') -> bool:
        """
        Function to verify whether the message should be serializable.

        :return: Boolean indicating if the message meets 
        the criteria for serialization.
        """
        if message["role"] in ["system", "tool"]:
            # Prevent serialization of messages
            # not intended for user display
            return False
        elif not message.get("content"):
            # Prevent serialization for messages
            # without meaningful content
            return False

        return True

    def _serialize_message(self, message: 'Conversation.Message'):
        return {
            "role": message["role"],
            "content": message["content"],
            "actions": message["actions"]
            }

    @property
    def serialized_messages(self) -> List['Conversation.Message']:
        """
        Returns a representation of the conversation, excluding system
        messages.

        :return: List of messages without system messages.
        """
        # Excluding system messages for privacy & security reasons
        serialized_messages = \
            [
                self._serialize_message(message)
                for message in self.messages
                if self._is_serialized(message)
            ]
        return serialized_messages


class Apps:
    def list(
            self,
            config: Optional[APIOptions] = None
            ) -> List[ApplicationListResponse]:
        """
        Lists all applications available to the user.

        :param config: Optional dictionary containing parameters
        for the list call.
        :return: List of applications.

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.
        """
        client = WriterAIManager.acquire_client()
        config = config or {}

        response = client.applications.list(**config)
        # Convert the response to list
        # to collect all the apps available
        result = list(response)

        return result

    def retrieve(
            self,
            application_id: str,
            config: Optional[APIOptions] = None
            ) -> ApplicationRetrieveResponse:
        """
        Retrieves all information about a specific application by its ID.

        :param application_id: The ID of the application to retrieve data for.
        :param config: Optional dictionary containing parameters
        for the retrieve call.
        :return: The application data.

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.
        """
        client = WriterAIManager.acquire_client()
        config = config or {}

        response_data = client.applications.retrieve(
            application_id=application_id,
            **config
            )
        return response_data

    def generate_content(
            self,
            application_id: str,
            input_dict: Optional[Dict[str, str]] = None,
            async_job: Optional[bool] = False,
            config: Optional[APIOptions] = None
            ) -> Union[str, JobCreateResponse]:
        """
        Generates output based on an existing AI Studio no-code application.

        :param application_id: The id for the application, which can be
            obtained on AI Studio.
        :param input_dict: Optional dictionary containing parameters for
            the generation call.
        :param async_job: Optional. If True, the function initiates
            an asynchronous job and returns job details instead of
            waiting for the immediate output.
        :return: The generated text
            or the information about new asynchronous job.
        :raises RuntimeError: If response data was not properly formatted
        to retrieve model text.

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.

        Examples:

        **Synchronous Call (Immediate Output)**

        >>> response = writer.ai.apps.generate_content(
        ...     application_id="app_123",
        ...     input_dict={"topic": "Climate Change"},
        ...     async_job=False
        ... )
        >>> print(response)
        "Climate change refers to long-term shifts in temperatures and weather patterns..."

        **Asynchronous Call (Job Creation)**

        >>> response = writer.ai.apps.generate_content(
        ...     application_id="app_123",
        ...     input_dict={"topic": "Climate Change"},
        ...     async_job=True
        ... )
        >>> print(response)
        JobCreateResponse(id="job_456", created_at=datetime(2025, 2, 24, 12, 30, 45), status="in_progress")
        >>> result = writer.ai.apps.retrieve_job(job_id=response.id)
        >>> if result.status == "completed":
        ...     print(result.data)
        {"title": "output", "suggestion": "Climate change refers to long-term shifts in temperatures and weather patterns..."}

        """

        client = WriterAIManager.acquire_client()
        config = config or {}
        input_dict = input_dict or {}
        inputs = []

        for k, v in input_dict.items():
            inputs.append(Input({
                "id": k,
                "value": v if isinstance(v, list) else [v]
            }))

        if not async_job:
            response_data = client.applications.generate_content(
                application_id=application_id,
                inputs=inputs,
                **config
                )

            text = response_data.suggestion
            if text:
                return text

            raise RuntimeError(
                "Failed to acquire proper response " +
                "for completion from data: " +
                f"{response_data}"
            )

        else:
            async_response_data = client.applications.jobs.create(
                application_id=application_id,
                inputs=inputs,
                **config
            )

            return async_response_data

    def retry_job(
            self,
            job_id: str,
            config: Optional[APIOptions] = None
            ) -> JobRetryResponse:
        """
        Retries a specific asynchronous job execution.

        :param job_id: The unique identifier of the job to retry.
        :param config: Optional API configuration options for the retry request.
        :return: The response data from retrying the job.

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.
        """
        client = WriterAIManager.acquire_client()
        config = config or {}

        response_data = client.applications.jobs.retry(
            job_id=job_id,
            **config
            )

        return response_data

    def retrieve_jobs(
            self,
            application_id: str,
            config: Optional[APIRetrieveJobsOptions] = None
            ) -> List[ApplicationGenerateAsyncResponse]:
        """
        Retrieves a list of jobs for a specific application.

        :param application_id: The unique identifier of the application.
        :param config: Optional configuration parameters for the API request.
        :return: A list of job responses associated with
        the specified application.

        The `config` dictionary can include:
        - `limit` (int): The pagination limit for retrieving the jobs.
        - `offset` (int): The pagination offset for retrieving the jobs.
        - `status` (Literal['in_progress', 'failed', 'completed']):
        Filter jobs by the provided status.
        - `extra_headers` (Optional[Headers]): Additional headers.
        - `extra_query` (Optional[Query]): Extra query parameters.
        - `extra_body` (Optional[Body]): Additional body data.
        - `timeout` (Union[float, Timeout, None, NotGiven]):
        Request timeout in seconds.
        """
        client = WriterAIManager.acquire_client()
        config = config or {}

        jobs = client.applications.jobs.list(
            application_id=application_id,
            **config
        )

        return jobs.result

    def retrieve_job(
            self,
            job_id: str,
            config: Optional[APIOptions] = None
            ) -> ApplicationGenerateAsyncResponse:
        """
        Retrieves an asynchronous job from the Writer AI API based on its ID.

        :param job_id: The unique identifier of the job to retrieve.
        :param config: Additional API configuration options.
        :returns: The retrieved job object from the API.

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.
        """
        client = WriterAIManager.acquire_client()
        config = config or {}

        job = client.applications.jobs.retrieve(
            job_id=job_id,
            **config
        )

        return job

    def retrieve_graphs(
            self,
            application_id: str,
            config: Optional[APIOptions] = None
            ) -> List[str]:
        """
        Retrieves a list of graph IDs for a specific application.

        :param application_id: The unique identifier of the application.
        :param config: Optional configuration parameters for the API request.
        :return: A list of graph IDs associated with the specified application.

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.
        """
        client = WriterAIManager.acquire_client()
        config = config or {}

        graphs = client.applications.graphs.list(
            application_id=application_id,
            **config
        )

        return graphs.graph_ids

    def associate_graphs(
            self,
            application_id: str,
            graph_ids: List[str],
            config: Optional[APIOptions] = None
            ) -> ApplicationGraphsResponse:
        """
        Associates a list of graph IDs with a specific application.

        :param application_id: The unique identifier of the application.
        :param graph_ids: A list of graph IDs to associate with the application.
        :param config: Optional configuration parameters for the API request.
        :return: The response data from associating the graphs.

        The `config` dictionary can include the following keys:
        - `extra_headers` (Optional[Headers]):
        Additional headers for the request.
        - `extra_query` (Optional[Query]):
        Additional query parameters for the request.
        - `extra_body` (Optional[Body]):
        Additional body parameters for the request.
        - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
        Timeout for the request in seconds.
        """
        client = WriterAIManager.acquire_client()
        config = config or {}

        response_data = client.applications.graphs.update(
            application_id=application_id,
            graph_ids=graph_ids,
            **config
            )

        return response_data


class Tools:
    SplittingStrategy = Union[
        Literal["llm_split"],
        Literal["fast_split"],
        Literal["hybrid_split"]
        ]
    MedicalResponseType = Union[
        Literal["Entities"],
        Literal["RxNorm"],
        Literal["ICD-10-CM"],
        Literal["SNOMED CT"]
    ]

    @staticmethod
    def _retrieve_tools_accessor():
        return WriterAIManager.acquire_client().tools

    @classmethod
    def parse_pdf(
        cls,
        file_id_or_file: Union[str, File, Dict],
        format: Union[Literal['text'], Literal['markdown']] = 'text',
        config: Optional[APIOptions] = None
    ) -> str:
        config = config or {}
        client_tools = cls._retrieve_tools_accessor()
        file_id = None
        if isinstance(file_id_or_file, File):
            file_id = file_id_or_file.id
        elif isinstance(file_id_or_file, Dict):
            if not (
                "data" in file_id_or_file
                and
                "type" in file_id_or_file
                and
                "name" in file_id_or_file
            ):
                raise ValueError(
                    "Invalid payload passed to parse_pdf method: " +
                    "expected keys `data`, `type` and `name`, " +
                    f"got {file_id_or_file.keys()}"
                    )
            new_file_from_payload = upload_file(
                **file_id_or_file,
                config=config
                )
            file_id = new_file_from_payload.id
        elif isinstance(file_id_or_file, str):
            file_id = file_id_or_file
        else:
            raise ValueError(
                "parse_pdf expects a `writer.ai.File` type instance, " +
                f"a file payload, or a string ID: got {type(file_id_or_file)}"
            )

        result = client_tools.parse_pdf(
            file_id=file_id,
            format=format,
            **config
        )

        return result.content

    @classmethod
    def split(
        cls,
        content: str,
        strategy: SplittingStrategy = "llm_split",
        config: Optional[APIOptions] = None
    ) -> List[str]:
        if not content:
            raise ValueError("Content cannot be empty.")
        config = config or {}
        client_tools = cls._retrieve_tools_accessor()

        result = client_tools.context_aware_splitting(
            strategy=strategy,
            text=content,
            **config
        )

        return result.chunks

    @classmethod
    def comprehend_medical(
        cls,
        content: str,
        response_type: MedicalResponseType = "Entities",
        config: Optional[APIOptions] = None
    ) -> List:
        if not content:
            raise ValueError("Content cannot be empty.")
        config = config or {}
        client_tools = cls._retrieve_tools_accessor()

        result = client_tools.comprehend.medical(
            content=content,
            response_type=response_type,
            **config
        )

        return result.entities


def complete(
        initial_text: str,
        config: Optional['CreateOptions'] = None
        ) -> str:
    """
    Completes the input text using the given data and returns the first
    resulting text choice.

    :param initial_text: The initial text prompt for the completion.
    :param config: Optional dictionary containing parameters
    for the completion call.
    :return: The text of the first choice from the completion response.
    :raises RuntimeError: If response data was not properly formatted
    to retrieve model text.
    """
    config = config or {}

    client = WriterAIManager.acquire_client()
    request_model = \
        config.get("model", None) or WriterAIManager.use_completion_model()

    response_data: Completion = client.completions.create(
        model=request_model,
        prompt=initial_text,
        best_of=config.get("best_of", NotGiven()),
        max_tokens=config.get("max_tokens", NotGiven()),
        random_seed=config.get("random_seed", NotGiven()),
        stop=config.get("stop", NotGiven()),
        temperature=config.get("temperature", NotGiven()),
        top_p=config.get("top_p", NotGiven()),
        extra_headers=config.get("extra_headers"),
        extra_body=config.get("extra_body"),
        extra_query=config.get("extra_query"),
        timeout=config.get("timeout")
        )

    for entry in response_data.choices:
        text = entry.text
        if text:
            return text

    raise RuntimeError(
        "Failed to acquire proper response for completion from data: " +
        f"{response_data}")


def stream_complete(
        initial_text: str,
        config: Optional['CreateOptions'] = None
        ) -> Generator[str, None, None]:
    """
    Streams completion results from an initial text prompt, yielding
    each piece of text as it is received.

    :param initial_text: The initial text prompt for the stream completion.
    :param config: Optional dictionary containing parameters
    for the stream completion call.
    :yields: Each text completion as it arrives from the stream.
    """
    config = config or {}

    client = WriterAIManager.acquire_client()
    request_model = \
        config.get("model", None) or WriterAIManager.use_completion_model()

    response: Stream = client.completions.create(
        model=request_model,
        prompt=initial_text,
        stream=True,
        best_of=config.get("best_of", NotGiven()),
        max_tokens=config.get("max_tokens", NotGiven()),
        random_seed=config.get("random_seed", NotGiven()),
        stop=config.get("stop", NotGiven()),
        temperature=config.get("temperature", NotGiven()),
        top_p=config.get("top_p", NotGiven()),
        extra_headers=config.get("extra_headers"),
        extra_body=config.get("extra_body"),
        extra_query=config.get("extra_query"),
        timeout=config.get("timeout")
        )

    for line in response:
        processed_line = _process_completion_data_chunk(line)
        if processed_line:
            yield processed_line
    else:
        response.close()


def _gather_graph_ids(graphs_or_graph_ids: list) -> List[str]:
    graph_ids = []
    for item in graphs_or_graph_ids:
        if isinstance(item, Graph):
            graph_ids.append(item.id)
        elif isinstance(item, str):
            graph_ids.append(item)
        else:
            raise ValueError(
                f"Invalid item in graphs_or_graph_ids list: {type(item)}"
                )

    return graph_ids


def ask(
    question: str,
    graphs_or_graph_ids: List[Union[Graph, str]],
    subqueries: bool = False,
    config: Optional[APIOptions] = None
):
    """
    Sends a question to the specified graph(s) and retrieves 
    a single response.

    :param question: The query or question to be answered by the graph(s).
    :param graphs_or_graph_ids: A list of `Graph` objects or graph IDs that
    should be queried.
    :param subqueries: Enables subquery generation if set to True,
    enhancing the result.
    :param config: Optional dictionary for additional API
    configuration settings.

    :return: The answer to the question from the graph(s).

    :raises ValueError: If an invalid graph or graph ID is provided
    in `graphs_or_graph_ids`.
    :raises RuntimeError: If the API response is improperly formatted
    or the answer cannot be retrieved.

    The `config` dictionary can include the following keys:
    - `extra_headers` (Optional[Headers]):
    Additional headers for the request.
    - `extra_query` (Optional[Query]):
    Additional query parameters for the request.
    - `extra_body` (Optional[Body]):
    Additional body parameters for the request.
    - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
    Timeout for the request in seconds.

    **Example Usage**:

    >>> response = ask(
    ...     question="What is the capital of France?",
    ...     graphs_or_graph_ids=["graph_id_1", "graph_id_2"]
    ... )
    """
    config = config or {}
    client = WriterAIManager.acquire_client()
    graph_ids = _gather_graph_ids(graphs_or_graph_ids)

    response = cast(
        Question,
        client.graphs.question(
            graph_ids=graph_ids,
            question=question,
            stream=False,
            subqueries=subqueries,
            **config
        )
    )

    return response.answer


def stream_ask(
    question: str,
    graphs_or_graph_ids: List[Union[Graph, str]],
    subqueries: bool = False,
    config: Optional[APIOptions] = None
) -> Generator[str, None, None]:
    """
    Streams response for a question posed to the specified graph(s).

    This method returns incremental chunks of the response, ideal for long 
    responses or when reduced latency is needed.

    :param question: The query or question to be answered by the graph(s).
    :param graphs_or_graph_ids: A list of Graph objects or graph IDs that
    should be queried.
    :param subqueries: Enables subquery generation if set to True,
    enhancing the result.
    :param config: Optional dictionary for additional API
    configuration settings.

    :yields: Incremental chunks of the answer to the question.

    :raises ValueError: If an invalid graph or graph ID
    is provided in `graphs_or_graph_ids`.

    The `config` dictionary can include the following keys:
    - `extra_headers` (Optional[Headers]):
    Additional headers for the request.
    - `extra_query` (Optional[Query]):
    Additional query parameters for the request.
    - `extra_body` (Optional[Body]):
    Additional body parameters for the request.
    - `timeout` (Union[float, httpx.Timeout, None, NotGiven]):
    Timeout for the request in seconds.

    **Example Usage**:

    >>> for chunk in stream_ask(
    ...     question="What are the benefits of renewable energy?",
    ...     graphs_or_graph_ids=["graph_id_1"]
    ... ):
    ...     print(chunk)
    ...
    """
    config = config or {}
    client = WriterAIManager.acquire_client()
    graph_ids = _gather_graph_ids(graphs_or_graph_ids)

    response = cast(
        Stream[QuestionResponseChunk],
        client.graphs.question(
            graph_ids=graph_ids,
            question=question,
            stream=True,
            subqueries=subqueries,
            **config
        )
    )

    for chunk in response._iter_events():
        raw_data = chunk.data
        answer = ""
        try:
            data = json.loads(raw_data)
            answer = data.get("answer", "")
        except json.JSONDecodeError:
            logging.error(
                "Couldn't parse chunk data during `question` streaming"
                )
        yield answer


def init(token: Optional[str] = None):
    """
    Initializes the WriterAIManager with an optional token.

    :param token: Optional token for authentication.
    :return: An instance of WriterAIManager.
    """
    return WriterAIManager(token=token)


apps = Apps()
tools = Tools()



import datetime
import statistics

import numpy as np
import pandas as pd
import plotly.express as px
import writer as wf
from writer.core import WriterState

# EVENT HANDLERS

def handle_timer_tick(state: WriterState):
    df = state["random_df"]
    for i in range(5):
        df[f'pgcf_{i+1}'] = np.around(np.random.rand(10, 1), decimals=9)
    state["random_df"] = df


def update(state, session):
    main_df = _get_main_df()
    main_df = main_df[main_df['length_cm'] >= state["filter"]["min_length"]]
    main_df = main_df[main_df['weight_g'] >= state["filter"]["min_weight"]]
    state['main_df'] = main_df

    paginated_members = _get_paginated_members(state['paginated_members_page'] - 1, state['paginated_members_page_size'])
    state['paginated_members'] = paginated_members
    state["session"] = session
    _update_metrics(state)
    _update_role_chart(state)
    _update_scatter_chart(state)


def handle_story_download(state):
    data = wf.pack_file("assets/story.txt", "text/plain")
    file_name = "hacker_pigeons.txt"
    state.file_download(data, file_name)


def handle_paginated_members_page_change(state, payload):
    page = payload
    maxpage = int(state["paginated_members_total_items"] / state["paginated_members_page_size"]) + 1
    if page > maxpage:
        state["paginated_members_page"] = maxpage - 2
    else:
        state["paginated_members_page"] = page

    update(state, None)


def handle_paginated_members_page_size_change(state, payload):
    state['paginated_members_page_size'] = payload
    update(state, None)


# LOAD / GENERATE DATA


def _generate_random_df():
    data = np.around(np.random.rand(10, 5), decimals=9)
    column_names = [f'pgcf_{i+1}' for i in range(5)]
    random_df = pd.DataFrame(data, columns=column_names)
    return random_df


def _get_main_df():
    main_df = pd.read_csv("assets/main_df.csv")
    return main_df


def _get_editable_df():

    df = pd.DataFrame({
        'number': [1, 2, 3],
        'boolean': [True, False, True],
        'object': [{"updatedAt": None}, {"updatedAt": None}, {"updatedAt": None}],
        'text': ['one', 'two', 'three'],
    })
    return wf.EditableDataFrame(df)

def _get_highlighted_members():
    sample_df = _get_main_df().sample(3).set_index("name", drop=False)
    sample = sample_df.to_dict("index")
    return sample

def _get_paginated_members(offset: int, limit: int):
    paginated_df = _get_main_df()[offset:offset + limit].set_index("name", drop=False)
    paginated = paginated_df.to_dict("index")
    return paginated

def _get_story_text():
    with open("assets/story.txt", "r") as f:
        return f.read()

# UPDATES

def _update_metrics(state):
    main_df = state["main_df"]
    bmi_list = (main_df['weight_g'] / 1000) / \
        ((main_df['length_cm'] / 100) ** 2)
    bmi = statistics.mean(bmi_list)
    diversity_index = main_df['feather_color'].nunique()/len(main_df)
    metrics = {
        "average_weight": round(main_df['weight_g'].mean(), 0),
        "average_length": round(main_df['length_cm'].mean(), 0),
        "average_bmi": round(bmi, 2),
        "diversity": round(diversity_index, 2),
    }
    metrics.update({
        "average_weight_note": "+Acceptable",
        "average_length_note": "+Acceptable",
        "average_bmi_note": "-Overweight" if metrics["average_bmi"] >= 5.2 else "+Acceptable",
        "diversity_note": "-Not diverse" if metrics["diversity"] < 0.8 else "+Acceptable",
    })
    state["metrics"] = metrics


def _update_role_chart(state):
    main_df = state["main_df"]
    role_counts = main_df['role'].value_counts().reset_index()
    role_counts.columns = ['role', 'count']
    fig = px.bar(role_counts, x='role', y='count')
    fig.update_layout(
        margin={"l": 20, "r": 20, "t": 20, "b": 50},
        showlegend=False
    )
    state["role_chart"] = fig


def _update_scatter_chart(state):
    main_df = state["main_df"]
    average_role_data = main_df.groupby("role").agg(
        {"length_cm": "mean", "weight_g": "mean"}).reset_index()
    fig = px.scatter(average_role_data, x="length_cm", y="weight_g", height=400,
        size_max=10, size="weight_g")
    fig.update_layout(
        margin={"l": 20, "r": 20, "t": 20, "b": 50},
        showlegend=False
    )
    state["scatter_chart"] = fig


# UPDATES DATAFRAME

# Subscribe this event handler to the `wf-dataframe-add` event
def on_editable_df_record_add(state, payload):
    state['editable_df'].record_add(payload)


# Subscribe this event handler to the `wf-dataframe-update` event
def on_editable_df_record_change(state, payload):
    payload['record']['object']['updatedAt'] = str(datetime.datetime.now())
    state['editable_df'].record_update(payload)

# Subscribe this event handler to the `wf-dataframe-action` event
def on_editable_df_record_action(state, payload):
    record_index = payload['record_index']
    if payload['action'] == 'remove':
        state['editable_df'].record_remove(payload)
    if payload['action'] == 'open':
        state['editable_df_open_text'] = str(state['editable_df'].record(record_index)['text'])
    if payload['action'] == 'notify':
        state.add_notification("info", "Test", "Notify on this row : " + str(state['editable_df'].record(record_index)))


# STATE INIT

initial_state = wf.init_state({
    "main_df": _get_main_df(),
    "editable_df": _get_editable_df(),
    "editable_df_open_text": "<none>",
    "highlighted_members": _get_highlighted_members(),
    "random_df": _generate_random_df(),
    "hue_rotation": 180,
    "paginated_members": _get_paginated_members(0, 2),
    "paginated_members_page": 1,
    "paginated_members_total_items": len(_get_main_df()),
    "paginated_members_page_size": 2,
    "story": {
        "text": _get_story_text(),  # For display
    },
    "filter": {
        "min_length": 25,
        "min_weight": 300,
    },
    "metrics": {}
})

update(initial_state, None)



from writer.blocks.addtostatelist import AddToStateList
from writer.blocks.calleventhandler import CallEventHandler
from writer.blocks.changepage import ChangePage
from writer.blocks.code import CodeBlock
from writer.blocks.foreach import ForEach
from writer.blocks.httprequest import HTTPRequest
from writer.blocks.logmessage import LogMessage
from writer.blocks.parsejson import ParseJSON
from writer.blocks.returnvalue import ReturnValue
from writer.blocks.runworkflow import RunWorkflow
from writer.blocks.setstate import SetState
from writer.blocks.uieventtrigger import UIEventTrigger
from writer.blocks.writeraddchatmessage import WriterAddChatMessage
from writer.blocks.writeraddtokg import WriterAddToKG
from writer.blocks.writerchat import WriterChat
from writer.blocks.writerclassification import WriterClassification
from writer.blocks.writercompletion import WriterCompletion
from writer.blocks.writerinitchat import WriterInitChat
from writer.blocks.writernocodeapp import WriterNoCodeApp

SetState.register("workflows_setstate")
WriterClassification.register("workflows_writerclassification")
WriterCompletion.register("workflows_writercompletion")
HTTPRequest.register("workflows_httprequest")
RunWorkflow.register("workflows_runworkflow")
WriterNoCodeApp.register("workflows_writernocodeapp")
ForEach.register("workflows_foreach")
LogMessage.register("workflows_logmessage")
WriterChat.register("workflows_writerchat")
WriterAddChatMessage.register("workflows_writeraddchatmessage")
ParseJSON.register("workflows_parsejson")
CallEventHandler.register("workflows_calleventhandler")
AddToStateList.register("workflows_addtostatelist")
ReturnValue.register("workflows_returnvalue")
WriterInitChat.register("workflows_writerinitchat")
WriterAddToKG.register("workflows_writeraddtokg")
UIEventTrigger.register("workflows_uieventtrigger")
CodeBlock.register("workflows_code")
ChangePage.register("workflows_changepage")



from typing import TYPE_CHECKING, Any, Dict, Type

import writer.evaluator
from writer.ss_types import WriterConfigurationError

if TYPE_CHECKING:
    from writer.core_ui import Component
    from writer.ss_types import InstancePath
    from writer.workflows import WorkflowRunner

WorkflowBlock_T = Type["WorkflowBlock"]
block_map:Dict[str, WorkflowBlock_T] = {}

class WorkflowBlock:
    
    @classmethod
    def register(cls, type: str):
        block_map[type] = cls

    def __init__(self, component_id: str, runner: "WorkflowRunner", execution_environment: Dict):
        self.outcome = None
        self.message = None
        self.component_id = component_id
        self.runner = runner
        self.execution_time_in_seconds = -1.0
        self.execution_environment = execution_environment
        self.result = None
        self.return_value = None
        self.instance_path: InstancePath = [{"componentId": component_id, "instanceNumber": 0}]
        self.evaluator = writer.evaluator.Evaluator(runner.session.session_state, runner.session.session_component_tree)

    def _handle_missing_field(self, field_key):
        component_tree = self.runner.session.session_component_tree
        component = component_tree.get_component(self.component_id)
        field_content = component.content.get(field_key)
        if field_content:
            raise WriterConfigurationError(f'The field `{field_key}` is required. The expression specified, `{field_content}`, resulted in an empty value.')
        else:
            raise WriterConfigurationError(f'The field `{field_key}` is required. It was left empty.')

    def _get_field(self, field_key: str, as_json=False, default_field_value=None, required=False):
        if default_field_value is None:
            if as_json:
                default_field_value = "{}"
            else:
                default_field_value = ""
        value = self.evaluator.evaluate_field(self.instance_path, field_key, as_json, default_field_value, self.execution_environment)

        if required and (value is None or value == "" or value == {}):
            self._handle_missing_field(field_key)
            
        return value

    def _set_state(self, expr: str, value: Any):
        self.evaluator.set_state(expr, self.instance_path, value, base_context=self.execution_environment)

    def run(self):
        pass


from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate


class WriterNoCodeApp(WorkflowBlock):
    @classmethod
    def register(cls, type: str):
        super(WriterNoCodeApp, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="workflows_node",
                writer={
                    "name": "No-code agent",
                    "description": "Run a no-code agent.",
                    "category": "Writer",
                    "fields": {
                        "appId": {
                            "name": "App Id",
                            "type": "App Id",
                            "desc": "The agent id can be found in the agent's URL. It has a UUID format.",
                            "validator": {
                                "type": "string",
                                "format": "uuid",
                            },
                        },
                        "appInputs": {"name": "App inputs", "type": "Key-Value", "default": "{}"},
                    },
                    "outs": {
                        "success": {
                            "name": "Success",
                            "description": "If the execution was successful.",
                            "style": "success",
                        },
                        "error": {
                            "name": "Error",
                            "description": "If the function raises an Exception.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def run(self):
        try:
            import writer.ai

            application_id = self._get_field("appId", required=True)
            app_inputs = self._get_field("appInputs", as_json=True)

            # config = {}
            # if model_id:
            #     config["model"] = model_id

            result = writer.ai.apps.generate_content(application_id, app_inputs).strip()
            self.result = result
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate, WriterConfigurationError


class ReturnValue(WorkflowBlock):

    @classmethod
    def register(cls, type: str):
        super(ReturnValue, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="workflows_node",
            writer={
                "name": "Return value",
                "description": "Returns a value from a workflow or sub-workflow.",
                "category": "Logic",
                "fields": {
                    "value": {
                        "name": "Value",
                        "type": "Text",
                        "control": "Textarea"
                    },
                },
                "outs": {
                    "success": {
                        "name": "Success",
                        "description": "If the function doesn't raise an Exception.",
                        "style": "success",
                    },
                    "error": {
                        "name": "Error",
                        "description": "If the function raises an Exception.",
                        "style": "error",
                    },
                },
            }
        ))

    def run(self):
        try:
            value = self._get_field("value")
            if value is None:
                raise WriterConfigurationError("Return value cannot be empty or None.")
            self.result = value
            self.return_value = value
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e


import plotly.graph_objects as go
import writer as wf
from sklearn.datasets import make_blobs
from sklearn.linear_model import LogisticRegression

COLOR = {
    0: '#3c64fa',
    1: '#00eba8',
    2: '#5a677c',
    3: '#ff8866',
    4: '#d4b2f7',
    5: '#c3e6ff',
    6: '#045758',
    7: '#001435',
    8: '#ec3d10',
    9: '#38006a'
}

def _line(x0, coef, intercept, c):
        return (-(x0 * coef[c, 0]) - intercept[c]) / coef[c, 1]

def update(state):
    cluster_std = state['cluster_std']
    multi_class = state['multi_class']
    number_of_points = int(state['number_of_points'])
    number_of_groups = int(state['number_of_groups'])
    X, y = make_blobs(
        n_samples=number_of_points,
        n_features=2,
        cluster_std=cluster_std,
        centers=number_of_groups
    )

    clf = LogisticRegression(
        solver="sag",
        max_iter=1000,
        random_state=42,
        multi_class=multi_class
    ).fit(X, y)

    coef = clf.coef_
    intercept = clf.intercept_
    score = clf.score(X, y)

    state["message"] = "training score : %.3f (%s)" % (score, multi_class)

    data = []
    for i in range(number_of_groups):
        data.append(
            go.Scatter(
                x=X[y == i][:, 0],
                y=X[y==i][:, 1],
                mode='markers',
                name='Group '+str(i),
                hoverinfo='none',
                marker={
                    'color': COLOR[i],
                    'symbol': 'circle',
                    'size': 10
                }
            )
        )

    for i in range(1 if number_of_groups < 3 else number_of_groups):
        data.append(go.Scatter(
            x=[-20, 20],
            y=[
                _line(-20, coef, intercept, i),
                _line(20, coef, intercept, i)
            ],
            mode='lines', 
            line={'color': COLOR[i], 'width': 2},
            name='Logistic Regression'
        ))

    layout = go.Layout(
        width=700,height=700,
        hovermode='closest', hoverdistance=1,
        xaxis={
            'title': 'Feature 1',
            'range': [-20,20],
            'fixedrange': True,
            'constrain': "domain",
            'scaleanchor': "y",
            'scaleratio': 1
        },
        yaxis={
            'title': 'Feature 2',
            'range': [-20,20],
            'fixedrange': True,
            'constrain': "domain"
        },
        paper_bgcolor='#FFFFFF',
        margin={'l': 30, 'r': 30, 't': 30, 'b': 30},
    )

    fig = go.Figure(data=data, layout=layout)
    state['figure'] = fig


initial_state = wf.init_state({
    "my_app": {
        "title": "Logistic regression visualizer"
    },
    "message": None,
    "figure": None,
    "multi_class": "ovr",
    "number_of_groups": 2,
    "number_of_points": 50,
    "cluster_std": 2,
})

update(initial_state)



import writer as wf
import writer.ai
from prompts import base_prompts, seo_keywords, user_prompt

# This is the base template for the Product Description Generator tutorial.
# More documentation is available at https://dev.writer.com/framework

# Initialize state here
wf.init_state({
    "form": {
        "title": "",
        "description": "",
        "keywords": ""
    },
    "message": "Fill in the inputs and click \"Generate\" to get started.",
})




from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate


class LogMessage(WorkflowBlock):
    @classmethod
    def register(cls, type: str):
        super(LogMessage, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="workflows_node",
                writer={
                    "name": "Log message",
                    "description": "Appends a message to the log.",
                    "category": "Other",
                    "fields": {
                        "type": {
                            "name": "Type",
                            "type": "Text",
                            "options": {"info": "Info", "error": "Error"},
                            "default": "info",
                        },
                        "message": {"name": "Message", "type": "Text", "control": "Textarea"},
                    },
                    "outs": {
                        "success": {
                            "name": "Success",
                            "description": "The request was successful.",
                            "style": "success",
                        },
                        "error": {
                            "name": "Error",
                            "description": "The workflow was executed successfully.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def run(self):
        try:
            type = self._get_field("type", False, "info")
            message = str(self._get_field("message", required=True))

            self.runner.session.session_state.add_log_entry(type, "Workflows message", message)
            self.result = message
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



from writer.abstract import register_abstract_template
from writer.blocks.base_block import WorkflowBlock
from writer.ss_types import AbstractTemplate

DEFAULT_MODEL = "palmyra-x-004"

class WriterCompletion(WorkflowBlock):

    @classmethod
    def register(cls, type: str):
        super(WriterCompletion, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="workflows_node",
            writer={
                "name": "Completion",
                "description": "Handles text completions.",
                "category": "Writer",
                "fields": {
                    "prompt": {
                        "name": "Prompt",
                        "type": "Text",
                        "control": "Textarea"
                    },
                    "modelId": {
                        "name": "Model id",
                        "type": "Text",
                        "default": DEFAULT_MODEL
                    },
                    "temperature": {
                        "name": "Temperature",
                        "type": "Number",
                        "default": "0.7",
                        "validator": {
                            "type": "number",
                            "minimum": 0,
                            "maximum": 1,
                        }
                    }
                },
                "outs": {
                    "success": {
                        "name": "Success",
                        "description": "If the function doesn't raise an Exception.",
                        "style": "success",
                    },
                    "error": {
                        "name": "Error",
                        "description": "If the function raises an Exception.",
                        "style": "error",
                    },
                },
            }
        ))

    def run(self):
        try:
            import writer.ai

            prompt = self._get_field("prompt")
            temperature = float(self._get_field("temperature", False, "0.7"))
            model_id = self._get_field("modelId", False, default_field_value=DEFAULT_MODEL)
            config = { "temperature": temperature, "model": model_id}
            result = writer.ai.complete(prompt, config).strip()
            self.result = result
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e


from writer.abstract import register_abstract_template
from writer.blocks.base_trigger import WorkflowTrigger
from writer.ss_types import AbstractTemplate


class UIEventTrigger(WorkflowTrigger):
    @classmethod
    def register(cls, type: str):
        super(UIEventTrigger, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="workflows_node",
                writer={
                    "name": "UI Trigger",
                    "description": "Trigger the workflow when an UI event takes place.",
                    "category": "Triggers",
                    "fields": {
                        "refComponentId": {
                            "name": "Component Id",
                            "type": "Text",
                            "options": "uiComponentsWithEvents",
                            "desc": "The id of the component that will trigger this branch.",
                        },
                        "refEventType": {
                            "name": "Event type",
                            "type": "Text",
                            "desc": "The type of the event that will trigger this branch. For example, wf-click.",
                            "options": "eventTypes",
                        },
                    },
                    "outs": {
                        "trigger": {
                            "name": "Trigger",
                            "style": "success",
                        },
                    },
                },
            ),
        )

    def run(self):
        self.outcome = "trigger"



import asyncio
import base64
import html
import importlib.util
import io
import json
import logging
import mimetypes
import os
import os.path
import pathlib
import socket
import textwrap
import time
import typing
from contextlib import asynccontextmanager
from importlib.machinery import ModuleSpec
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union, cast
from urllib.parse import urlsplit

import uvicorn
from fastapi import FastAPI, HTTPException, Request, Response
from fastapi.responses import FileResponse, JSONResponse
from fastapi.routing import Mount
from fastapi.staticfiles import StaticFiles
from pydantic import ValidationError
from starlette.websockets import WebSocket, WebSocketDisconnect, WebSocketState

from writer import VERSION, abstract, crypto
from writer.ai import Graph
from writer.app_runner import AppRunner
from writer.ss_types import (
    AppProcessServerResponse,
    ComponentUpdateRequestPayload,
    EventResponsePayload,
    HashRequestPayload,
    HashRequestResponsePayload,
    InitRequestBody,
    InitResponseBodyEdit,
    InitResponseBodyRun,
    InitSessionRequestPayload,
    InitSessionResponsePayload,
    ServeMode,
    StateEnquiryResponsePayload,
    WriterEvent,
    WriterWebsocketIncoming,
    WriterWebsocketOutgoing,
)

if typing.TYPE_CHECKING:
    from .auth import Auth, Unauthorized

MAX_WEBSOCKET_MESSAGE_SIZE = 201 * 1024 * 1024
logging.getLogger().setLevel(logging.INFO)


class JobVault:
    SCHEMES: List[str] = []
    job_vault_implementations: List[Type["JobVault"]] = []

    def __init__(self):
        self.counter = 0
        self.vault = {}

    def generate_job_id(self):
        self.counter += 1
        return str(self.counter)

    def set(self, job_id: str, value: Any):
        self.vault[job_id] = value

    def get(self, job_id: str):
        return self.vault.get(job_id)

    @classmethod
    def register(cls, klass: Type["JobVault"]):
        cls.job_vault_implementations.insert(0, klass)

    @classmethod
    def _get_matching_implementation(cls, connection_string):
        for job_vault_implementation in cls.job_vault_implementations:
            for scheme in job_vault_implementation.SCHEMES:
                if connection_string.startswith(scheme):
                    return job_vault_implementation

    @classmethod
    def create_vault(cls):
        connection_string = os.getenv("WRITER_PERSISTENT_STORE")
        if not connection_string:
            return cls()

        matching_implementation = cls._get_matching_implementation(connection_string)
        if not matching_implementation:
            supported_schemes = [
                scheme
                for implementation in JobVault.job_vault_implementations
                for scheme in implementation.SCHEMES
            ]
            supported_schemes_msg = ", ".join(supported_schemes)
            logging.error(f"No matching implementation found for { connection_string }. Falling back to in-memory JobVault. \
                          Supported schemes: {supported_schemes_msg}.")
            return cls()

        try:
            return matching_implementation()
        except Exception as e:
            logging.error(
                f"There was an error connecting to { connection_string }. Falling back to in-memory JobVault. {repr(e)}"
            )
            return cls()


class RedisJobVault(JobVault):
    SCHEMES = ["redis://", "rediss://", "redis-socket://", "redis-sentinel://"]
    DEFAULT_TTL = 86400

    def __init__(self):
        import redis  # type: ignore

        super().__init__()
        redis_connection_string = os.getenv("WRITER_PERSISTENT_STORE")
        self.redis_client = redis.from_url(
            redis_connection_string, decode_responses=True, socket_timeout=30
        )
        self.counter_key = "job_counter"
        if not self.redis_client.exists(self.counter_key):
            self.redis_client.set(self.counter_key, 0)

    def generate_job_id(self):
        job_id = self.redis_client.incr(self.counter_key)
        return str(job_id)

    def set(self, job_id: str, value: Any):
        ttl = RedisJobVault.DEFAULT_TTL
        env_ttl = os.getenv("WRITER_PERSISTENT_STORE_TTL")
        if env_ttl is not None:
            ttl = int(env_ttl)
        json_str = json.dumps(value)
        self.redis_client.set(f"job:{job_id}", json_str, ex=ttl)

    def get(self, job_id: str):
        json_str = self.redis_client.get(f"job:{job_id}")
        if not json_str:
            return None
        return json.loads(json_str)


class WriterState(typing.Protocol):
    app_runner: AppRunner
    writer_app: bool
    job_vault: JobVault
    is_server_static_mounted: bool
    meta: Union[Dict[str, Any], Callable[[], Dict[str, Any]]]  # meta tags for SEO
    opengraph_tags: Union[
        Dict[str, Any], Callable[[], Dict[str, Any]]
    ]  # opengraph tags for social networks integration (facebook, discord)
    title: Union[str, Callable[[], str]]  # title of the page, default: "Writer Framework"


class WriterAsgi(typing.Protocol):
    state: WriterState


class WriterFastAPI(FastAPI, WriterAsgi):  # type: ignore
    pass


app: WriterFastAPI = cast(WriterFastAPI, None)


def get_asgi_app(
    user_app_path: str,
    serve_mode: ServeMode,
    enable_remote_edit: bool = False,
    enable_server_setup: bool = True,
    on_load: Optional[Callable] = None,
    on_shutdown: Optional[Callable] = None,
    enable_jobs_api: bool = False,
) -> WriterFastAPI:
    """
    Builds an ASGI server that can be injected into another ASGI application
    or an asgi server like uvicorn

    >>> asgi_app = writer.serve.get_asgi_app("app1", "run")
    >>> uvicorn.run(asgi_app, host="0.0.0.0", port=5328)

    :param user_app_path: writer application path
    :param serve_mode: server mode (run, edit)
    :param enable_remote_edit: allow editing from the internet (by default, editing only works locally)
    :param enable_server_setup: enables fastapi setup hook on startup, server_setup.py
    :param on_load: callback called on loading
    :param on_shutdown: callback called at shutdown
    :return: ASGI Server
    """
    global app
    if serve_mode not in ["run", "edit"]:
        raise ValueError("""Invalid mode. Must be either "run" or "edit".""")

    _fix_mimetype()
    app_runner = AppRunner(user_app_path, serve_mode)

    @asynccontextmanager
    async def lifespan(asgi_app: FastAPI):
        nonlocal app_runner

        app_runner.hook_to_running_event_loop()
        app_runner.load()

        if (
            on_load is not None
            and hasattr(asgi_app.state, "is_server_static_mounted")
            and asgi_app.state.is_server_static_mounted
        ):
            on_load()

        try:
            yield
        except asyncio.CancelledError:
            pass

        app_runner.shut_down()
        if on_shutdown is not None:
            on_shutdown()

    app = cast(WriterFastAPI, FastAPI(lifespan=lifespan))
    """
    Reuse the same pattern to give variable to FastAPI application
    than `app.state.is_server_static_mounted` already use in Writer Framework.
    """
    app.state.writer_app = True
    app.state.app_runner = app_runner

    def _get_extension_paths() -> List[str]:
        extensions_path = pathlib.Path(user_app_path) / "extensions"
        if not extensions_path.exists():
            return []
        filtered_files = [
            f
            for f in extensions_path.rglob("*")
            if f.suffix.lower() in (".js", ".css") and f.is_file()
        ]
        relative_paths = [f.relative_to(extensions_path).as_posix() for f in filtered_files]
        return relative_paths

    cached_extension_paths = _get_extension_paths()

    def _check_origin_header(origin_header: Optional[str]) -> bool:
        if serve_mode not in ("edit") or enable_remote_edit is True:
            return True
        if origin_header is None:
            return False
        hostname = urlsplit(origin_header).hostname
        if hostname in ("127.0.0.1", "localhost"):
            return True
        return False

    # Init

    def _get_run_starter_pack(payload: InitSessionResponsePayload):
        return InitResponseBodyRun(
            mode="run",
            sessionId=payload.sessionId,
            userState=payload.userState,
            mail=payload.mail,
            components=payload.components,
            userFunctions=payload.userFunctions,
            extensionPaths=cached_extension_paths,
            featureFlags=payload.featureFlags,
            abstractTemplates=abstract.templates,
        )

    def _get_edit_starter_pack(payload: InitSessionResponsePayload):
        run_code: Optional[str] = app_runner.run_code

        return InitResponseBodyEdit(
            mode="edit",
            sessionId=payload.sessionId,
            userState=payload.userState,
            mail=payload.mail,
            components=payload.components,
            userFunctions=payload.userFunctions,
            runCode=run_code,
            sourceFiles=app_runner.source_files,
            extensionPaths=cached_extension_paths,
            featureFlags=payload.featureFlags,
            abstractTemplates=abstract.templates,
        )

    @app.get("/api/health")
    async def health():
        return {"status": "ok"}

    @app.post("/api/init")
    async def init(
        initBody: InitRequestBody, request: Request, response: Response
    ) -> Union[InitResponseBodyRun, InitResponseBodyEdit]:
        """
        Handles session init and provides a "starter pack" to the frontend.
        """

        origin_header = request.headers.get("origin")
        if not _check_origin_header(origin_header):
            wrong_origin_message = "A session request with origin %s was rejected. For security reasons, only local origins are allowed in edit mode. "
            wrong_origin_message += "To circumvent this protection, use the --enable-remote-edit flag if running via command line."
            logging.error(wrong_origin_message, origin_header)
            raise HTTPException(
                status_code=403, detail="Incorrect origin. Only local origins are allowed."
            )

        session_id = request.cookies.get("session")
        if session_id is not None:
            initBody.proposedSessionId = session_id

        app_response = await app_runner.init_session(
            InitSessionRequestPayload(
                cookies=dict(request.cookies),
                headers=dict(request.headers),
                proposedSessionId=initBody.proposedSessionId,
            )
        )

        status = app_response.status

        """
        Deletes the session cookie that was set by 
        authentication when it is present.
        """
        if session_id is not None:
            response.delete_cookie("session")

        if status == "error" or app_response.payload is None:
            raise HTTPException(status_code=403, detail="Session rejected.")

        if serve_mode == "run":
            return _get_run_starter_pack(app_response.payload)

        if serve_mode == "edit":
            return _get_edit_starter_pack(app_response.payload)

    # Jobs

    async def _get_payload_as_json(request: Request):
        payload = None
        body = await request.body()
        if not body:
            return None
        try:
            payload = await request.json()
        except json.JSONDecodeError:
            raise HTTPException(status_code=400, detail="Cannot parse the payload.")
        return payload

    @app.post("/api/job/workflow/{workflow_key}")
    async def create_workflow_job(workflow_key: str, request: Request, response: Response):
        if not enable_jobs_api:
            raise HTTPException(status_code=404)

        crypto.verify_message_authorization_signature(f"create_job_{workflow_key}", request)

        def serialize_result(data):
            if isinstance(data, list):
                return [serialize_result(item) for item in data]
            if isinstance(data, dict):
                return {k: serialize_result(v) for k, v in data.items()}
            if isinstance(data, (str, int, float, bool, type(None))):
                return data
            try:
                return json.loads(json.dumps(data))
            except (TypeError, OverflowError):
                return f"Can't be displayed. Value of type: {str(type(data))}."

        def update_job(job_id: str, job_info: dict):
            current_job_info = app.state.job_vault.get(job_id)
            if not current_job_info:
                raise RuntimeError("Job not found.")
            merged_info = current_job_info | {"finished_at": int(time.time())} | job_info
            app.state.job_vault.set(job_id, merged_info)

        def job_done_callback(task: asyncio.Task, job_id: str):
            try:
                apsr: Optional[AppProcessServerResponse] = task.result()
                if apsr is None or apsr.status != "ok":
                    update_job(job_id, {"status": "error"})
                    return
                result = None
                if apsr.payload and apsr.payload.result:
                    result = apsr.payload.result.get("result")
                update_job(job_id, {"status": "complete", "result": serialize_result(result)})
            except Exception as e:
                update_job(job_id, {"status": "error"})
                raise e

        app_response = await app_runner.init_session(
            InitSessionRequestPayload(
                cookies=dict(request.cookies), headers=dict(request.headers), proposedSessionId=None
            )
        )

        if not app_response or not app_response.payload:
            raise HTTPException(status_code=500, detail="Cannot initialize session.")
        session_id = app_response.payload.sessionId
        is_session_ok = await app_runner.check_session(session_id)
        if not is_session_ok:
            raise HTTPException(status_code=500, detail="Cannot initialize session.")

        loop = asyncio.get_running_loop()
        task = loop.create_task(
            app_runner.handle_event(
                session_id,
                WriterEvent(
                    type="wf-builtin-run",
                    isSafe=True,
                    handler=f"$runWorkflow_{workflow_key}",
                    payload=await _get_payload_as_json(request),
                ),
            )
        )

        job_id = app.state.job_vault.generate_job_id()
        app.state.job_vault.set(
            job_id, {"id": job_id, "status": "in progress", "created_at": int(time.time())}
        )
        task.add_done_callback(lambda t: job_done_callback(t, job_id))
        return {"id": job_id, "token": crypto.get_hash(f"get_job_{job_id}")}

    @app.get("/api/job/{job_id}")
    async def get_workflow_job(job_id: str, request: Request, response: Response):
        if not enable_jobs_api:
            raise HTTPException(status_code=404)

        crypto.verify_message_authorization_signature(f"get_job_{job_id}", request)
        job = app.state.job_vault.get(job_id)

        if not job:
            return JSONResponse(status_code=404, content={"id": job_id, "status": "not found"})

        status_code = 200
        if job.get("status") == "error":
            status_code = 400

        return JSONResponse(status_code=status_code, content=job)

    # Streaming

    async def _stream_session_init(websocket: WebSocket):
        """
        Waits for the client to provide a session id to initialise the stream.
        Returns the session id received.
        """

        session_id = None
        while session_id is None:
            req_message_raw = await websocket.receive_json()

            try:
                req_message = WriterWebsocketIncoming.model_validate(req_message_raw)
            except ValidationError:
                logging.error("Incorrect incoming request.")
                return

            if req_message.type == "streamInit" and req_message.payload is not None:
                session_id = req_message.payload.get("sessionId")
        return session_id

    async def _stream_incoming_requests(websocket: WebSocket, session_id: str):
        """
        Handles incoming requests from client.
        """

        pending_tasks: Set[asyncio.Task] = set()

        try:
            while True:
                req_message_raw = await websocket.receive_json()

                try:
                    req_message = WriterWebsocketIncoming.model_validate(req_message_raw)
                except ValidationError:
                    logging.error("Incorrect incoming request.")
                    break

                is_session_ok = await app_runner.check_session(session_id)
                if not is_session_ok:
                    break

                new_task = None

                if req_message.type == "event":
                    new_task = asyncio.create_task(
                        _handle_incoming_event(websocket, session_id, req_message)
                    )
                elif req_message.type == "keepAlive":
                    new_task = asyncio.create_task(
                        _handle_keep_alive_message(websocket, session_id, req_message)
                    )
                elif req_message.type == "stateEnquiry":
                    new_task = asyncio.create_task(
                        _handle_state_enquiry_message(websocket, session_id, req_message)
                    )
                elif serve_mode == "edit" and req_message.type == "hashRequest":
                    new_task = asyncio.create_task(
                        _handle_hash_request(websocket, session_id, req_message)
                    )
                elif serve_mode == "edit":
                    new_task = asyncio.create_task(
                        _handle_incoming_edit_message(websocket, session_id, req_message)
                    )

                if new_task:
                    pending_tasks.add(new_task)
                    new_task.add_done_callback(pending_tasks.discard)
        except WebSocketDisconnect:
            pass
        except asyncio.CancelledError:
            raise
        finally:
            # Cancel pending tasks

            for pending_task in pending_tasks.copy():
                pending_task.cancel()
                try:
                    await pending_task
                except asyncio.CancelledError:
                    pass

    async def _handle_incoming_event(
        websocket: WebSocket, session_id: str, req_message: WriterWebsocketIncoming
    ):
        response = WriterWebsocketOutgoing(
            messageType=f"{req_message.type}Response",
            trackingId=req_message.trackingId,
            payload=None,
        )

        # Allows for global events if in edit mode (such as "Run workflow" for previewing a workflow)

        is_safe = serve_mode == "edit"
        res_payload: Optional[Dict[str, Any]] = None
        apsr: Optional[AppProcessServerResponse] = None
        apsr = await app_runner.handle_event(
            session_id,
            WriterEvent(
                type=req_message.payload.get("type"),
                handler=req_message.payload.get("handler"),
                isSafe=is_safe,
                instancePath=req_message.payload.get("instancePath"),
                payload=req_message.payload.get("payload"),
            ),
        )
        if apsr is not None and apsr.payload is not None:
            res_payload = typing.cast(EventResponsePayload, apsr.payload).model_dump()
        if res_payload is not None:
            response.payload = res_payload
        await websocket.send_json(response.model_dump())

    async def _handle_incoming_edit_message(
        websocket: WebSocket, session_id: str, req_message: WriterWebsocketIncoming
    ):
        response = WriterWebsocketOutgoing(
            messageType=f"{req_message.type}Response",
            trackingId=req_message.trackingId,
            payload=None,
        )
        if req_message.type == "componentUpdate":
            await app_runner.update_components(
                session_id,
                ComponentUpdateRequestPayload(components=req_message.payload["components"]),
            )
        elif req_message.type == "codeSaveRequest":
            app_runner.save_code(
                session_id, req_message.payload["code"], req_message.payload["path"]
            )
        elif req_message.type == "codeUpdate":
            app_runner.update_code(session_id, req_message.payload["code"])
        elif req_message.type == "loadSourceFile":
            path = os.path.join(*req_message.payload["path"])
            try:
                response.payload = {"content": app_runner.load_persisted_script(path)}
            except FileNotFoundError as error:
                logging.warning(f"could not load script at {path}", error)
                response.payload = {"error": str(error)}
        elif req_message.type == "createSourceFile":
            path = os.path.join(*req_message.payload["path"])
            try:
                app_runner.create_persisted_script(path)
            except Exception as error:
                response.payload = {"error": str(error)}
        elif req_message.type == "deleteSourceFile":
            path = os.path.join(*req_message.payload["path"])
            try:
                app_runner.delete_persisted_script(path)
            except Exception as error:
                response.payload = {"error": str(error)}
        elif req_message.type == "renameSourceFile":
            from_path = os.path.join(*req_message.payload["from"])
            to_path = os.path.join(*req_message.payload["to"])
            try:
                app_runner.rename_persisted_script(from_path, to_path)
            except Exception as error:
                response.payload = {"error": str(error)}
        elif req_message.type == "listResources":
            res = await app_runner.list_resources(session_id, req_message.payload["resource_type"])
            response.payload = res.payload
        elif req_message.type == "uploadSourceFile":
            path = os.path.join(*req_message.payload["path"])

            try:
                content = base64.b64decode(req_message.payload["content"])
                app_runner.create_persisted_script(path, content)
                response.payload = {"sourceFiles": app_runner.source_files}
            except Exception as error:
                response.payload = {"error": str(error)}

        await websocket.send_json(response.model_dump())

    async def _handle_keep_alive_message(
        websocket: WebSocket, session_id: str, req_message: WriterWebsocketIncoming
    ):
        response = WriterWebsocketOutgoing(
            messageType="keepAliveResponse", trackingId=req_message.trackingId, payload=None
        )
        await websocket.send_json(response.model_dump())

    async def _handle_state_enquiry_message(
        websocket: WebSocket, session_id: str, req_message: WriterWebsocketIncoming
    ):
        response = WriterWebsocketOutgoing(
            messageType=f"{req_message.type}Response",
            trackingId=req_message.trackingId,
            payload=None,
        )
        res_payload: Optional[Dict[str, Any]] = None
        apsr: Optional[AppProcessServerResponse] = None
        apsr = await app_runner.handle_state_enquiry(session_id)
        if apsr is not None and apsr.payload is not None:
            res_payload = typing.cast(StateEnquiryResponsePayload, apsr.payload).model_dump()
        if res_payload is not None:
            response.payload = res_payload
        await websocket.send_json(response.model_dump())

    async def _handle_hash_request(
        websocket: WebSocket, session_id: str, req_message: WriterWebsocketIncoming
    ):
        response = WriterWebsocketOutgoing(
            messageType=f"{req_message.type}Response",
            trackingId=req_message.trackingId,
            payload=None,
        )
        apsr: Optional[AppProcessServerResponse] = None
        apsr = await app_runner.handle_hash_request(
            session_id, HashRequestPayload(message=req_message.payload.get("message", ""))
        )
        if apsr is not None and apsr.payload is not None:
            response.payload = typing.cast(HashRequestResponsePayload, apsr.payload).model_dump()
        await websocket.send_json(response.model_dump())

    async def _stream_outgoing_announcements(websocket: WebSocket, session_id: str):
        """
        Handles outgoing communications to the client (announcements).
        """

        session_queue: asyncio.Queue = asyncio.Queue()
        app_runner.announcement_queues[session_id] = session_queue

        try:
            while True:
                announcement_data = await session_queue.get()
                announcement = WriterWebsocketOutgoing(
                    messageType="announcement", trackingId=-1, payload=announcement_data
                )
                await websocket.send_json(announcement.dict())
                if announcement_data.get("type") == "codeUpdate":
                    return
        except WebSocketDisconnect:
            pass
        finally:
            del app_runner.announcement_queues[session_id]

    @app.websocket("/api/stream")
    async def stream(websocket: WebSocket):
        """Initialises incoming and outgoing communications on the stream."""

        await websocket.accept()

        origin_header = websocket.headers.get("origin")
        if not _check_origin_header(origin_header):
            await websocket.close(code=1008)
            return

        try:
            session_id = await _stream_session_init(websocket)
        except WebSocketDisconnect:
            return

        is_session_ok = await app_runner.check_session(session_id)
        if not is_session_ok:
            await websocket.close(code=1008)  # Invalid permissions
            return

        task1 = asyncio.create_task(_stream_incoming_requests(websocket, session_id))
        task2 = asyncio.create_task(_stream_outgoing_announcements(websocket, session_id))

        try:
            await asyncio.wait((task1, task2), return_when=asyncio.FIRST_COMPLETED)
            await asyncio.sleep(1)
            task1.cancel()
            task2.cancel()
            await task1
            await task2
        except asyncio.CancelledError:
            pass

    # Mount static paths

    user_app_static_path = pathlib.Path(user_app_path) / "static"
    if user_app_static_path.exists():
        app.mount("/static", StaticFiles(directory=str(user_app_static_path)), name="user_static")

    user_app_extensions_path = pathlib.Path(user_app_path) / "extensions"
    if user_app_extensions_path.exists():
        app.mount(
            "/extensions", StaticFiles(directory=str(user_app_extensions_path)), name="extensions"
        )

    server_path = pathlib.Path(__file__)
    server_static_path = server_path.parent / "static"
    if server_static_path.exists():
        _mount_render_index_html(app, server_static_path)
        _mount_server_static_path(app, server_static_path)
        app.state.is_server_static_mounted = True
    else:
        logging.error(
            textwrap.dedent(
                """\
                \x1b[31;20mError: Failed to acquire server static path. Writer Framework may not be properly built.

                To resolve this issue, try the following steps:
                1. Run the 'npm run build' script in the 'ui' directory and then restart the app.
                2. Alternatively, launch a UI instance by running 'npm run dev' in the 'ui' directory.

                Please refer to the CONTRIBUTING.md for detailed instructions.\x1b[0m"""
            )
        )

    JobVault.register(RedisJobVault)

    # Return
    if enable_server_setup is True:
        _execute_server_setup_hook(user_app_path)

    app.state.job_vault = JobVault.create_vault()

    return app


def print_init_message():
    print(f"""               
                                                               
                   &@@@@@@@@@@     ,@@@@@@@@@@*     @@@@@@@@@@                  
                   .@@@@@@@@@@(     &@@@@@@@@@@     *@@@@@@@@*                  
                    %@@@@@@@@@@     .@@@@@@@@@@(     @@@@@@@@                   
                     @@@@@@@@@@&     #@@@@@@@@@@      @@@@@@                    
                     ,@@@@@@@@@@,     @@@@@@@@@@@     (@@@@(                    
                      &@@@@@@@@@@     .@@@@@@@@@@*     @@@@                     
                       @@@@@@@@@@%     %@@@@@@@@@@     .@@,                     
                       (@@@@@@@@@@.     @@@@@@@@@@%     %&                      
                        @@@@@@@@@@&     (@@@@@@@@@@      .                      
                        *@@@@@@@@@@,     @@@@@@@@@@&                            
                         @@@@@@@@@@@     ,@@@@@@@@@@*                           
                          @@@@@@@@@@#     %@@@@@@@@@@ 


WRITER FRAMEWORK v{VERSION}""")


def print_route_message(run_name: str, port: int, host: str):
    GREEN_TOKEN = "\033[92m"
    END_TOKEN = "\033[0m"

    print(
        f"{run_name} is available at:{END_TOKEN}{GREEN_TOKEN} http://{host}:{port}{END_TOKEN}",
        flush=True,
    )


def register_auth(
    auth: "Auth",
    callback: Optional[Callable[[Request, str, dict], None]] = None,
    unauthorized_action: Optional[Callable[[Request, "Unauthorized"], Response]] = None,
):
    auth.register(app, callback=callback, unauthorized_action=unauthorized_action)


def serve(
    app_path: str,
    mode: ServeMode,
    port: Optional[int],
    host,
    enable_remote_edit=False,
    enable_server_setup=False,
    enable_jobs_api=False,
):
    """Initialises the web server."""

    print_init_message()

    def on_load():
        run_name = "Builder" if mode == "edit" else "App"
        print_route_message(run_name, port, host)

    """
    Loading of the server_setup.py is active by default 
    when Writer Framework is launched with the run command.
    """
    if port is None:
        mode_allowed_ports = {"run": (3005, 3099), "edit": (4005, 4099)}

        port = _next_localhost_available_port(mode_allowed_ports[mode])

    enable_server_setup = mode == "run" or enable_server_setup
    app = get_asgi_app(
        app_path,
        mode,
        enable_remote_edit,
        on_load=on_load,
        enable_server_setup=enable_server_setup,
        enable_jobs_api=enable_jobs_api,
    )
    log_level = "warning"
    uvicorn.run(
        app, host=host, port=port, log_level=log_level, ws_max_size=MAX_WEBSOCKET_MESSAGE_SIZE
    )


@asynccontextmanager
async def lifespan(app: FastAPI):
    """
    This feature supports launching multiple Writer Framework applications simultaneously.

    >>> import uvicorn
    >>> import writer.serve
    >>> from fastapi import FastAPI, Response
    >>>
    >>> root_asgi_app = FastAPI(lifespan=writer.serve.lifespan)
    >>>
    >>> sub_asgi_app_1 = writer.serve.get_asgi_app("../app1", "run")
    >>> sub_asgi_app_2 = writer.serve.get_asgi_app("../app2", "run")
    >>>
    >>> uvicorn.run(root_asgi_app, ws_max_size=writer.serve.MAX_WEBSOCKET_MESSAGE_SIZE)

    Writer Framework uses lifespan to start an application server (app_runner) per
    application.
    """
    writer_lifespans = []
    for route in app.routes:
        if isinstance(route, Mount) and isinstance(route.app, FastAPI):
            if hasattr(route.app.state, "writer_app"):
                ctx = route.app.router.lifespan_context
                writer_lifespans.append(ctx)

    async with _lifespan_invoke(writer_lifespans, app):
        yield


def configure_webpage_metadata(
    title: Union[str, Callable[[], str]] = "Writer Framework",
    meta: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]] = None,
    opengraph_tags: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]] = None,
):
    """
    Configures the page header for SEO and social networks from `server_setup` module.

    >>> writer.serve.configure_webpage_metadata(
    >>>     title="my App",
    >>>     meta={
    >>>         "description": "my amazing app",
    >>>         "keywords": "WF, Amazing, AI App",
    >>>         "author": "Amazing company"
    >>>     }
    >>>)

    Meta will accept description, keywords, author. Other meta tags as view port won't be supported.

    Settings accept functions to adapt content based on application data.

    >>> def generated_title():
    >>>     return "My App" # load title using info from database

    >>> def generated_meta_tags():
    >>>     {
    >>>         "description": "my amazing app",
    >>>         "keywords": "WF, Amazing, AI App",
    >>>         "author": "Amazing company"
    >>>     }

    >>> writer.serve.configure_webpage_metadata(
    >>>     title=generated_title
    >>>     meta=generated_meta_tags
    >>> )

    OpenGraph tags are used by social networks to display information about the page. WF support them.

    >>> writer.serve.configure_webpage_metadata(
    >>>     title=generated_title
    >>>     opengraph_tags= {
    >>>         "og:title": "My App",
    >>>         "og:description": "My amazing app",
    >>>         "og:image": "https://myapp.com/logo.png",
    >>>         "og:url": "https://myapp.com"
    >>>     }
    >>> )

    >>> def generated_opengraph_tags():
    >>>     return {
    >>>         "og:title": "My App",
    >>>         "og:description": "My amazing app",
    >>>     }

    >>> writer.serve.configure_webpage_metadata(
    >>>     title=generated_title
    >>>     opengraph_tags= generated_opengraph_tags
    >>> )

    ---

    WF replaces the placeholders <!-- {{ meta }} -->, <!-- {{ opengraph_tags }} -->
    and the <title>Writer framework<title> tag in the index.html file.

    :param title: The title of the page. Default: "Writer Framework"
    :param meta: A list of meta tags. Default: {}
    :param opengraph_tags: A dictionary of OpenGraph tags. Default: {}
    """
    app.state.title = title
    app.state.meta = meta if meta is not None else {}
    app.state.opengraph_tags = opengraph_tags if opengraph_tags is not None else {}


@asynccontextmanager
async def _lifespan_invoke(context: list, app: FastAPI):
    """
    Helper to run multiple lifespans in cascade.

    Running

    >>> _lifespan_invoke([app1.router.lifespan_context, app2.router.lifespan_context], app)

    is equivalent to

    >>> @asynccontextmanager
    >>> async def lifespan_context(app: FastAPI):
    >>>   async with app1.router.lifespan_context(app):
    >>>     async with app2.router.lifespan_context(app):
    >>>       yield
    """
    ctx = context.pop(0)
    async with ctx(app):
        if len(context) > 0:
            async with _lifespan_invoke(context, app):
                yield
        else:
            yield


def _fix_mimetype():
    """
    Fixes mimetypes for .js files. This is needed for the webserver to serve .js files correctly.
    """
    js_mimetype = mimetypes.guess_type("myfile.js")[0]
    if js_mimetype[0] != "text/javascript":
        mimetypes.add_type("text/javascript", ".js")


def _mount_server_static_path(app: FastAPI, server_static_path: pathlib.Path) -> None:
    """
    Unitarily declares the files and folders present in "/static" directory of source code.

    We avoid the general declaration as below. This declaration limit the ability of a developper to
    declare it's own route.

    >>> asgi_app.mount("/", StaticFiles(directory=str(server_static_path), html=True), name="server_static")

    Writer Framework routes remain priority. A developer cannot come and overload them.
    """
    for f in wf_root_static_assets():
        if f.is_file():
            app.get(f"/{f.name}")(lambda: FileResponse(f))
        if f.is_dir():
            app.mount(f"/{f.name}", StaticFiles(directory=f), name=f"server_static_{f}")


def _mount_render_index_html(app: FastAPI, server_static_path: pathlib.Path):
    """
    Serves the main page with the title that has been configured.

    :param app:
    :param server_static_path:
    :return:
    """

    def _render_index_html():
        with io.open(server_static_path.joinpath("index.html"), "r", encoding="utf-8") as f:
            index_html = f.read()
            if hasattr(app.state, "title"):
                index_html = index_html.replace(
                    "<title>Writer Framework</title>",
                    f"<title>{html.escape(app.state.title)}</title>",
                )

            if hasattr(app.state, "meta"):
                meta = app.state.meta() if callable(app.state.meta) else app.state.meta
                meta_tags = "\n".join(
                    [f'<meta name="{k}" content="{html.escape(v)}">' for k, v in meta.items()]
                )
                index_html = index_html.replace("<!-- {{ meta }} -->", meta_tags)
            else:
                index_html = index_html.replace("<!-- {{ meta }} -->", "")

            if hasattr(app.state, "opengraph_tags"):
                opengraph_tags = (
                    app.state.opengraph_tags()
                    if callable(app.state.opengraph_tags)
                    else app.state.opengraph_tags
                )
                opengraph_tags = "\n".join(
                    [
                        f'<meta property="{k}" content="{html.escape(v)}">'
                        for k, v in opengraph_tags.items()
                    ]
                )
                index_html = index_html.replace("<!-- {{ opengraph_tags }} -->", opengraph_tags)
            else:
                index_html = index_html.replace("<!-- {{ opengraph_tags }} -->", "")

        return Response(content=index_html, media_type="text/html")

    return app.get("/")(_render_index_html)


def app_runner(asgi_app: WriterFastAPI) -> AppRunner:
    return asgi_app.state.app_runner


def wf_root_static_assets() -> List[pathlib.Path]:
    """
    Lists the root writer Framework static assets. Some of them are files, some other are directories.

    >>> for f in wf_root_static_assets()
    >>>     print(f"{f.name}")
    >>>     # favicon.ico
    >>>     # assets

    """
    all_static_assets: List[pathlib.Path] = []
    server_path = pathlib.Path(__file__)
    server_static_path = server_path.parent / "static"
    for f in server_static_path.glob("*"):
        all_static_assets.append(f)

    return all_static_assets


def _execute_server_setup_hook(user_app_path: str) -> None:
    """
    Runs the server_setup.py module if present in the application directory.

    """
    server_setup_path = os.path.join(user_app_path, "server_setup.py")
    if os.path.isfile(server_setup_path):
        spec = cast(
            ModuleSpec, importlib.util.spec_from_file_location("server_setup", server_setup_path)
        )
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)  # type: ignore


def _next_localhost_available_port(port_range: Tuple[int, int]) -> int:
    """
    Searches for a free port in a given range on localhost to start the server

    >>> port = _next_localhost_available_port((3005, 3099))

    3005 is the first port to be tested. If it is not available, the port 3006 is tested, and so on.
    """
    for port in range(port_range[0], port_range[1]):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(1)
        result = sock.connect_ex(("127.0.0.1", port))
        sock.close()
        if result != 0:
            return port

    raise OSError(
        f"No free port found to start the server between {port_range[0]} and {port_range[1]} ."
    )



import os

import pandas as pd
import pytest


class TestApp:
    @pytest.fixture(autouse=True)
    def before(self):
        os.chdir(os.path.dirname(__file__))
        import main
        self.app_state = main.initial_state
        yield

    def test_handle_timer_tick(self):
        import main
        main.handle_timer_tick(self.app_state)
        df1 = self.app_state["random_df"].to_dict()
        main.handle_timer_tick(self.app_state)
        df2 = self.app_state["random_df"].to_dict()
        assert df1 != df2

    def test_metrics(self):
        import main
        data = {
            "weight_g": [3000, 3500, 3200, 3100, 2900, 3300],
            "length_cm": [50, 52, 51, 48, 47, 53],
            "feather_color": ["blue", "blue", "red", "green", "blue", "red"],
        }

        main_df = pd.DataFrame(data)
        self.app_state["main_df"] = main_df

        main._update_metrics(self.app_state)

        assert self.app_state["metrics"]["average_weight"] == 3167
        assert self.app_state["metrics"]["average_length"] == 50
        assert self.app_state["metrics"]["average_bmi"] == 12.6
        assert self.app_state["metrics"]["diversity"] == 0.5

        assert self.app_state["metrics"]["average_weight_note"] == "+Acceptable"
        assert self.app_state["metrics"]["average_length_note"] == "+Acceptable"
        assert self.app_state["metrics"]["average_bmi_note"] == "-Overweight"
        assert self.app_state["metrics"]["diversity_note"] == "-Not diverse"
